/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::fs::File;
use std::path::{Path, PathBuf};

use clap::Parser;
use env_logger::Env;

use anyhow::bail;
use bare_metrics_recorder::recording::BareMetricsRecorderCore;
use chrono::{DateTime, Local, NaiveDate, NaiveDateTime, TimeZone, Utc};
use datman::commands::backup::{backup_all_sources_to_destination, backup_source_to_destination};
use datman::commands::ilabel::interactive_labelling_session;
use datman::commands::init_descriptor;
use datman::descriptor::{load_descriptor, SourceDescriptor};
use datman::get_hostname;
use datman::remote::backup_source_requester::backup_remote_source_to_destination;
use datman::remote::backup_source_responder;
use indicatif::{ProgressBar, ProgressDrawTarget, ProgressStyle};
use log::info;
use std::str::FromStr;

pub const FAILURE_SYMBOL_OBNOXIOUS_FLASHING: &str = "\x1b[5m\x1b[31m⚠️ \x1b[25m\x1b[22m";
pub const BOLD: &str = "\x1b[1m";
pub const BOLD_OFF: &str = "\x1b[22m";
pub const WHITE: &str = "\x1b[37m";
pub const RED: &str = "\x1b[31m";
pub const GREEN: &str = "\x1b[32m";

#[derive(Parser)]
pub enum DatmanCommand {
    /// Initialise a datman descriptor in this directory.
    Init {},

    ///
    Status {},

    #[clap(name = "ilabel")]
    InteractiveLabelling {
        /// Name of the source to label.
        source_name: String,
    },

    #[clap(name = "ibrowse")]
    InteractiveBrowsing {
        /// Name of the source to browse.
        source_name: String,
    },

    /// Back up a source locally or over the network.
    BackupOne {
        /// Name of the source to back up.
        source_name: String,

        /// Name of the destination to back up to.
        destination_name: String,
    },

    BackupAll {
        /// Name of the remote to back up.
        /// Special value 'self' means 'this host only'.
        /// Special value 'all' means 'all hosts'.
        remote_name: String,

        /// Name of the destination to back up to.
        destination_name: String,
    },

    Extract {
        /// Name of the 'source' to extract
        /// Omit for 'all'.
        #[clap(short)]
        source_name: Option<String>,

        /// If specified, will get the first backup after this date.
        #[clap(long)]
        after: Option<HumanDateTime>,

        /// If specified, will get the last backup before this date. The default behaviour is to get the latest.
        #[clap(long)]
        before: Option<HumanDateTime>,

        /// If not specified, time-restricted extractions that don't have a pointer for every source
        /// will instead lead to an error.
        #[clap(long)]
        accept_partial: bool, // TODO unimplemented.

        /// Name of the pile to extract from
        pile_name: String,

        /// Place to extract to.
        destination: PathBuf,

        /// Skip applying metadata. Might be needed to extract without superuser privileges.
        #[clap(long)]
        skip_metadata: bool,
    },

    #[clap(name = "_backup_source_responder")]
    InternalBackupSourceResponder,
}

pub struct HumanDateTime(pub DateTime<Local>);

impl FromStr for HumanDateTime {
    type Err = anyhow::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if let Ok(date_only) = NaiveDate::parse_from_str(s, "%Y-%m-%d") {
            let local_date = chrono::offset::Local.from_local_date(&date_only).unwrap();
            let local_datetime = local_date.and_hms(0, 0, 0);
            Ok(HumanDateTime(local_datetime))
        } else if let Ok(date_and_time) = NaiveDateTime::parse_from_str(s, "%Y-%m-%dT%H:%M:%S") {
            let local_datetime = chrono::offset::Local
                .from_local_datetime(&date_and_time)
                .unwrap();
            Ok(HumanDateTime(local_datetime))
        } else if let Ok(date_and_time) = NaiveDateTime::parse_from_str(s, "%Y-%m-%d %H:%M:%S") {
            let local_datetime = chrono::offset::Local
                .from_local_datetime(&date_and_time)
                .unwrap();
            Ok(HumanDateTime(local_datetime))
        } else {
            bail!("Couldn't parse using either format. Use one of: 2021-05-16 OR 2021-05-16T17:42:14 OR 2021-05-16 17:42:14");
        }
    }
}

fn with_obvious_successfail_message<R>(result: anyhow::Result<R>) -> anyhow::Result<R> {
    match &result {
        Ok(_) => {
            eprintln!("Operation {}successful{}.", GREEN, WHITE);
        }
        Err(error) => {
            eprintln!("{:?}", error);
            eprintln!(
                "{}{}Operation {}{}FAILED{}!{}",
                FAILURE_SYMBOL_OBNOXIOUS_FLASHING, WHITE, RED, BOLD, WHITE, BOLD_OFF
            );
        }
    };
    result
}

fn with_exitcode<R>(result: anyhow::Result<R>) {
    match &result {
        Ok(_) => {
            std::process::exit(0);
        }
        Err(_) => {
            std::process::exit(5);
        }
    };
}

fn main() -> anyhow::Result<()> {
    env_logger::Builder::from_env(Env::default().default_filter_or("info")).init();

    let now = Utc::now();

    let (shard, stopper) = BareMetricsRecorderCore::new(File::create(format!(
        "/tmp/datman_{}.baremetrics",
        now.format("%F_%H%M%S")
    ))?)
    .start("datman".to_string())?;
    shard.install_as_metrics_recorder()?;

    let opts: DatmanCommand = DatmanCommand::parse();

    match opts {
        DatmanCommand::Init {} => {
            init_descriptor(Path::new(".")).unwrap();
        }
        DatmanCommand::Status { .. } => {
            unimplemented!();
        }
        DatmanCommand::InteractiveLabelling { source_name } => {
            interactive_labelling_session(Path::new("."), source_name).unwrap();
        }
        DatmanCommand::InteractiveBrowsing { source_name } => {
            datman::commands::ibrowse::session(Path::new("."), source_name).unwrap();
        }
        DatmanCommand::BackupOne {
            source_name,
            destination_name,
        } => {
            let my_hostname = get_hostname();
            let descriptor = load_descriptor(Path::new(".")).unwrap();
            let source = &descriptor.sources[&source_name];
            let destination = &descriptor.piles[&destination_name];

            let mut pbar = ProgressBar::with_draw_target(0, ProgressDrawTarget::stdout_with_hz(10));
            pbar.set_style(
                ProgressStyle::default_bar().template(
                    "[{elapsed_precise}]/[{eta}] {bar:40.cyan/blue} {pos:>7}/{len:7} {msg}",
                ),
            );
            pbar.set_message("storing");

            let is_remote = if let SourceDescriptor::DirectorySource { hostname, .. } = source {
                hostname != &my_hostname
            } else {
                false
            };

            let result = if is_remote {
                backup_remote_source_to_destination(
                    source,
                    destination,
                    &descriptor,
                    Path::new("."),
                    &source_name,
                    &destination_name,
                    yama::utils::get_number_of_workers("YAMA_CHUNKERS"),
                    pbar,
                )
            } else {
                backup_source_to_destination(
                    source,
                    destination,
                    &descriptor,
                    Path::new("."),
                    &source_name,
                    &destination_name,
                    yama::utils::get_number_of_workers("YAMA_CHUNKERS"),
                    &mut pbar,
                )
            };
            with_exitcode(with_obvious_successfail_message(result))
        }
        DatmanCommand::BackupAll {
            remote_name,
            destination_name,
        } => {
            let descriptor = load_descriptor(Path::new(".")).unwrap();
            let destination = &descriptor.piles[&destination_name];

            let mut pbar = ProgressBar::with_draw_target(0, ProgressDrawTarget::stdout_with_hz(10));
            pbar.set_style(
                ProgressStyle::default_bar().template(
                    "[{elapsed_precise}]/[{eta}] {bar:40.cyan/blue} {pos:>7}/{len:7} {msg}",
                ),
            );
            pbar.set_message("storing");

            backup_all_sources_to_destination(
                destination,
                &descriptor,
                Path::new("."),
                &destination_name,
                yama::utils::get_number_of_workers("YAMA_CHUNKERS"),
                &mut pbar,
                remote_name,
            )
            .unwrap();
        }
        DatmanCommand::Extract {
            source_name,
            after,
            before,
            accept_partial,
            pile_name,
            destination,
            skip_metadata,
        } => {
            if !accept_partial {
                bail!("Specify --accept-partial until running without it is supported.");
            }

            if after.is_some() && before.is_some() {
                bail!("Can't specify both before and after!");
            }

            let before = before.map(|dt| dt.0.with_timezone(&Utc));
            let after = after.map(|dt| dt.0.with_timezone(&Utc));

            datman::commands::extract::extract(
                &destination,
                Path::new("."),
                source_name.as_ref().map(|x| x.as_ref()),
                &pile_name,
                before.into(),
                after.into(),
                !skip_metadata,
                !skip_metadata,
                !skip_metadata,
                yama::utils::get_number_of_workers("YAMA_EXTRACTORS"),
            )?;
        }

        DatmanCommand::InternalBackupSourceResponder => {
            info!("Datman responder at {:?}", std::env::current_exe()?);
            backup_source_responder::handler_stdio()?;
        }
    }
    Ok(())
}
