use crate::commands::backup::{get_pointer_name_at, label_filter_and_convert};
use crate::descriptor::{Descriptor, DestPileDescriptor, SourceDescriptor};
use crate::tree::FileTree;
use anyhow::{anyhow, bail};
use chrono::Utc;
use log::info;
use std::io::{Read, Write};
use std::path::Path;
use std::process::{Child, Command, Stdio};
use std::sync::Arc;
use yama::commands::{load_pile_descriptor, open_pile};
use yama::definitions::TreeNode;
use yama::pile::{Pile, RawPile, StoragePipelineSettings};
use yama::progress::ProgressTracker;
use yama::remote::responder::{Responder, ResponderWritingPipeline};
use yama::remote::{read_message, write_message};
use yama::utils::get_number_of_workers;

// SECURITY WARNING: the system you connect to using this mechanism will receive full access to
// your Yama pile. Do NOT connect to untrusted or compromised systems using this mechanism (yet).

pub fn introduction<R: Read, W: Write>(read: &mut R, write: &mut W) -> anyhow::Result<()> {
    info!("Introduction.");
    let version = env!("CARGO_PKG_VERSION");
    write_message(
        write,
        &format!("Datman v{} Backup Source Requester", version),
    )?;
    write.flush()?;

    let foreign_side: String = read_message(read)?;
    let expected_foreign_side = format!("Datman v{} Backup Source Responder", version);
    if &foreign_side != &expected_foreign_side {
        bail!(
            "Datman version mismatch. Expected {:?}, got {:?}",
            expected_foreign_side,
            foreign_side
        );
    }

    Ok(())
}

pub fn scanning<R: Read, W: Write>(
    read: &mut R,
    write: &mut W,
    path: &Path,
) -> anyhow::Result<Option<FileTree<(), (), (), ()>>> {
    info!("Scanning.");
    write_message(write, &"scan")?;
    write_message(write, &path)?;
    write.flush()?;
    let scan_result: Option<FileTree<(), (), (), ()>> = read_message(read)?;

    Ok(scan_result)
}

pub fn chunking<
    R: Read + Send + 'static,
    W: Write + Send + 'static,
    RP: RawPile + 'static,
    PT: ProgressTracker + Send + 'static,
>(
    read: R,
    mut write: W,
    path: &Path,
    pointer_name: String,
    tree_node: &TreeNode,
    raw_pile: Arc<RP>,
    parent: Option<String>,
    progress_bar: PT,
    use_writing_pipeline: bool,
) -> anyhow::Result<(R, W)> {
    info!("Chunking.");
    write_message(&mut write, &"chunk")?;
    write_message(&mut write, &path)?;
    write_message(&mut write, &pointer_name)?;
    write_message(&mut write, tree_node)?;
    write_message(&mut write, &parent)?;
    write.flush()?;

    let (writing_pipeline, control_rx) = if use_writing_pipeline {
        let sps = StoragePipelineSettings {
            num_compressors: get_number_of_workers("YAMA_PL_COMPRESSORS") as u32,
            compressor_input_bound: 32,
            writer_input_bound: 32,
        };
        let (control_tx, control_rx) = crossbeam_channel::unbounded();
        let pipeline = raw_pile.build_storage_pipeline(sps, control_tx)?;
        (
            Some(ResponderWritingPipeline {
                pipeline_submission: pipeline,
            }),
            Some(control_rx),
        )
    } else {
        (None, None)
    };

    let (r_handle, w_handle, join_handles) = Responder::start(
        read,
        write,
        get_number_of_workers("YAMA_RESPONDERS") as u16,
        raw_pile,
        writing_pipeline,
        progress_bar,
    );

    info!("Waiting for remote to finish chunking.");

    for handle in join_handles {
        handle.join().expect("Join handle should not fail");
    }
    let read = r_handle.join().unwrap();
    let write = w_handle.join().unwrap();

    if let Some(control_rx) = control_rx {
        while let Ok(_) = control_rx.recv() {
            // TODO nop
        }
    }

    info!("Remote finished chunking.");

    Ok((read, write))
}

pub fn quit<R: Read, W: Write>(read: &mut R, write: &mut W) -> anyhow::Result<()> {
    write_message(write, &"exit")?;
    write.flush()?;
    let scan_result: String = read_message(read)?;
    if scan_result.as_str() != "exit" {
        bail!("Exit failed");
    }

    Ok(())
}

pub fn connect_to_remote(descriptor: &Descriptor, hostname: &str) -> anyhow::Result<Child> {
    let remote_host_descriptor = descriptor
        .remote_hosts
        .get(hostname)
        .ok_or_else(|| anyhow::anyhow!("No remote host by that name: {:?}.", hostname))?;

    let connection = Command::new("ssh")
        .arg(&remote_host_descriptor.user_at_host)
        .arg("--")
        .arg(
            &remote_host_descriptor
                .path_to_datman
                .as_ref()
                .map(|x| x.as_str())
                .unwrap_or("datman"),
        )
        .arg("_backup_source_responder")
        .stdin(Stdio::piped())
        .stdout(Stdio::piped())
        .stderr(Stdio::inherit())
        .spawn()?;

    Ok(connection)
}

pub fn backup_remote_source_to_destination<PT: ProgressTracker + Send + 'static>(
    source: &SourceDescriptor,
    dest: &DestPileDescriptor,
    descriptor: &Descriptor,
    desc_path: &Path,
    source_name: &str,
    dest_name: &str,
    _num_workers: u8,
    progress_bar: PT,
) -> anyhow::Result<()> {
    match source {
        SourceDescriptor::DirectorySource {
            hostname,
            directory,
        } => {
            let remote_host_descriptor = descriptor
                .remote_hosts
                .get(hostname)
                .ok_or_else(|| anyhow::anyhow!("No remote host by that name: {:?}.", hostname))?;

            info!(
                "Looking to backup {} (from {}) to {}",
                source_name, remote_host_descriptor.user_at_host, dest_name
            );

            let connection = connect_to_remote(descriptor, hostname)?;
            let mut read = connection.stdout.expect("Requested stdout");
            let mut write = connection.stdin.expect("Requested stdin");

            // first start off with an introduction
            info!("Connecting...");
            introduction(&mut read, &mut write)?;

            // then request to scan
            info!("Requesting scan... (this may take some time)");
            let scan_result = scanning(&mut read, &mut write, directory.as_ref())?
                .ok_or_else(|| anyhow!("Remote scan failed (does the directory exist?)"))?;

            let root =
                label_filter_and_convert(scan_result, descriptor, desc_path, source_name, dest)?
                    .ok_or_else(|| anyhow!("Empty filter..."))?;

            let absolute_dest_path = desc_path.join(&dest.path);
            let pile_descriptor = load_pile_descriptor(&absolute_dest_path)?;
            let pile = open_pile(&absolute_dest_path, &pile_descriptor)?;

            let pointer_name = get_pointer_name_at(&source_name, Utc::now());
            if pile.read_pointer(pointer_name.as_str())?.is_some() {
                bail!(
                    "Pointer by name {:?} already exists; refusing to overwrite.",
                    pointer_name
                );
            }
            info!("Will write as pointer {:?}.", pointer_name);

            info!("Searching for suitable parents.");
            let mut parent: Option<String> = None;
            let prefix = format!("{}+", source_name);
            for pointer in pile.list_pointers()?.iter() {
                if pointer.starts_with(&prefix) {
                    match parent.as_ref() {
                        None => {
                            parent = Some(pointer.to_owned());
                        }
                        Some(cur_parent) => {
                            if cur_parent < pointer {
                                parent = Some(pointer.to_owned());
                            }
                        }
                    }
                }
            }

            match parent.as_ref() {
                Some(parent) => {
                    info!("Using parent: {}", parent);
                }
                None => {
                    info!("No suitable parent found.");
                }
            }

            info!("Storing remote using Yama (this may take some time)...");

            let raw_pile = Arc::new(pile.raw_pile);
            let pile = Pile::new(raw_pile.clone());

            let (mut read, mut write) = chunking(
                read,
                write,
                directory.as_ref(),
                pointer_name.clone(),
                &root,
                raw_pile,
                parent,
                progress_bar,
                true,
            )?;

            quit(&mut read, &mut write)?;
            pile.flush()?;

            info!("Stored! Checking for existence...");
            if pile.list_pointers()?.contains(&pointer_name) {
                info!("Exists!");
            } else {
                bail!("Pointer {:?} does not exist...", &pointer_name);
            }
        }
        SourceDescriptor::VirtualSource { .. } => {
            unimplemented!("Can't currently back up virtualsources on remotes...")
        }
    }
    Ok(())
}
