// This file implements the responder side of the backup source protocol -- the protocol used
// to connect to remote backup sources.

use crate::tree::scan;
use anyhow::bail;
use crossbeam_channel::Sender;
use log::info;
use std::io::{stdin, stdout, Read, Write};
use std::path::PathBuf;
use std::sync::Arc;
use std::time::Instant;
use yama::definitions::TreeNode;
use yama::pile::{Pile, RawPile};
use yama::progress::ProgressTracker;
use yama::remote::requester::Requester;
use yama::remote::{read_message, write_message, RequestBody, ResponseBody};
use yama::utils::get_number_of_workers;

pub fn introduction<R: Read, W: Write>(read: &mut R, write: &mut W) -> anyhow::Result<()> {
    let version = env!("CARGO_PKG_VERSION");
    write_message(
        write,
        &format!("Datman v{} Backup Source Responder", version),
    )?;
    write.flush()?;

    let foreign_side: String = read_message(read)?;
    let expected_foreign_side = format!("Datman v{} Backup Source Requester", version);
    if &foreign_side != &expected_foreign_side {
        bail!(
            "Datman version mismatch. Expected {:?}, got {:?}",
            expected_foreign_side,
            foreign_side
        );
    }

    Ok(())
}

pub fn scanning<R: Read, W: Write>(read: &mut R, write: &mut W) -> anyhow::Result<()> {
    let path: PathBuf = read_message(read)?;
    let scan_result = scan(&path)?;
    write_message(write, &scan_result)?;
    write.flush()?;
    Ok(())
}

pub fn chunking<R: Read + Send + 'static, W: Write + Send + 'static>(
    mut read: R,
    write: W,
) -> anyhow::Result<()> {
    let path: PathBuf = read_message(&mut read)?;
    let pointer_name: String = read_message(&mut read)?;
    let tree_node: TreeNode = read_message(&mut read)?;
    let parent: Option<String> = read_message(&mut read)?;

    let (yama_requester, requester_join_handles) = Requester::new(read, write);

    let raw_pile: Box<dyn RawPile> = Box::new(yama_requester);

    let pile = Pile::new(raw_pile);

    // TODO TODO progress
    let progress_bar = &mut ();

    yama::operations::storing::store_fully(
        Arc::new(pile),
        &path,
        &pointer_name,
        tree_node,
        parent,
        get_number_of_workers("YAMA_CHUNKERS"),
        progress_bar,
        true,
    )?;

    for join_handle in requester_join_handles {
        join_handle.join().expect("Expected to join handle");
    }

    Ok(())
}

pub struct ProgressSender {
    pub last_sent: Instant,
    pub current_progress: u64,
    pub current_max: u64,
    // TODO actually propagate this
    pub current_message: String,
    pub sender: Sender<(RequestBody, Option<Sender<ResponseBody>>)>,
}

impl ProgressSender {
    pub fn send_now(&mut self, include_message: bool) {
        self.sender
            .send((
                RequestBody::Progress {
                    current: self.current_progress,
                    max: self.current_max,
                },
                None,
            ))
            .expect("Progress sender failed");
        self.last_sent = Instant::now();
    }

    pub fn send_if_overdue(&mut self) {
        //info!("send if overdue...");
        if Instant::now().duration_since(self.last_sent).as_millis() >= 1024 {
            self.send_now(false);
        }
    }
}

impl ProgressTracker for ProgressSender {
    fn inc_progress(&mut self, delta_progress: u64) {
        self.current_progress += delta_progress;
        self.send_if_overdue();
    }

    fn set_current(&mut self, current_progress: u64) {
        self.current_progress = current_progress;
        self.send_if_overdue();
    }

    fn set_max_size(&mut self, max_size: u64) {
        self.current_max = max_size;
        self.send_if_overdue();
    }
}

pub fn chunking_stdio() -> anyhow::Result<()> {
    let (path, pointer_name, tree_node, parent) = {
        let stdin = stdin();
        let mut read = stdin.lock();
        let path: PathBuf = read_message(&mut read)?;
        let pointer_name: String = read_message(&mut read)?;
        let tree_node: TreeNode = read_message(&mut read)?;
        let parent: Option<String> = read_message(&mut read)?;
        (path, pointer_name, tree_node, parent)
    };

    info!(
        "Have pointer_name = {:?}, parent = {:?}",
        pointer_name, parent
    );

    let requester_join_handles = {
        let (yama_requester, requester_join_handles) = Requester::new_from_stdio();
        let command_sender = yama_requester.clone_command_sender();
        info!("progress sender in use");
        let mut progress_bar = ProgressSender {
            last_sent: Instant::now(),
            current_progress: 0,
            current_max: 0,
            current_message: "".to_string(),
            sender: command_sender,
        };

        let raw_pile: Box<dyn RawPile> = Box::new(yama_requester);

        let pile = Pile::new(raw_pile);

        yama::operations::storing::store_fully(
            Arc::new(pile),
            &path,
            &pointer_name,
            tree_node,
            parent,
            get_number_of_workers("YAMA_CHUNKERS"),
            &mut progress_bar,
            true,
        )?;

        requester_join_handles
    };

    info!("Waiting to join.");

    for join_handle in requester_join_handles {
        join_handle.join().expect("Expected to join handle");
    }

    info!("Chunking completed.");

    Ok(())
}

pub fn handler<R: Read + Send + 'static, W: Write + Send + 'static>(
    mut read: R,
    mut write: W,
) -> anyhow::Result<()> {
    introduction(&mut read, &mut write)?;

    scanning(&mut read, &mut write)?;

    chunking(read, write)?;

    Ok(())
}

pub fn handler_stdio() -> anyhow::Result<()> {
    let stdin = stdin();
    let stdout = stdout();

    let mut read = stdin.lock();
    let mut write = stdout.lock();
    info!("Introduction.");
    introduction(&mut read, &mut write)?;

    loop {
        let command: String = read_message(&mut read)?;
        match command.as_str() {
            "scan" => {
                info!("Scanning.");
                scanning(&mut read, &mut write)?;
            }
            "chunk" => {
                info!("Chunking.");
                drop(read);
                drop(write);
                chunking_stdio()?;
                read = stdin.lock();
                write = stdout.lock();
            }
            "exit" => {
                write_message(&mut write, &"exit")?;
                write.flush()?;
                break;
            }
            othercommand => {
                bail!("Don't understand {:?}", othercommand);
            }
        }
    }

    Ok(())
}
