/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::path::{Path, PathBuf};

use anyhow::{bail, Context};
use clap::{crate_authors, crate_description, crate_version, Parser};
use log::info;

use env_logger::Env;
use std::sync::Arc;
use yama::commands::{fully_integrate_pointer_node, load_pile_descriptor, open_pile};
use yama::debug::{debug_command, DebugCommand};
use yama::operations::checking::VacuumMode;
use yama::operations::pushpull::{determine_bypass_level, open_pile_with_work_bypass, push_to};
use yama::operations::{checking, extracting};
use yama::pile::{Pile, PileDescriptor, RawPile};
use yama::{commands, debug};

#[derive(Parser)]
#[clap(version = crate_version!(), author = crate_authors!(), about = crate_description!())]
struct Opts {
    /// Chooses a different pile to be the working pile.
    /// If specified, must be the name of a remote in yama.toml.
    // TODO OBS? #[clap(short, long)]
    // with: Option<String>,

    #[clap(subcommand)]
    command: PileCommand,
}

#[derive(Parser)]
enum PileCommand {
    /// Initialise a yama pile in this directory.
    Init {},

    /// Retrieve a pointer from the yama pile, using a named pointer name.
    Retrieve {
        /// Name of the pointer to retrieve.
        pointer_name: String,

        /// Limited expression(s) of files to retrieve.
        #[clap(short, long)]
        subset: Vec<PathBuf>,

        destination: PathBuf,

        /// Number of extraction workers to use. Ideal value varies, but probably not much more than
        /// the number of CPU threads.
        #[clap(long)]
        num_workers: Option<u8>,
    },

    /// Check this yama pile for corruption.
    Check {
        #[clap(long)]
        apply_gc: bool,

        #[clap(long)]
        dry_run_gc: bool,

        #[clap(long)]
        deep: bool,

        #[clap(long)]
        shallow: bool,
    },

    /// Enter a debug prompt for manually operating on the yama pile.
    Debug { supplied_command: Vec<String> },

    /// Pushes a pointer from this pile to another pile.
    Push {
        /// The name of the pointer to push.
        pointer_name: String,

        /// The path to the other pile to push the pointer to.
        other_pile_path: PathBuf,
    },
}

fn main() -> anyhow::Result<()> {
    std::process::exit(wrapped_main()?);
}

/// Wrapped main, returning the exit code.
/// DO NOT USE exit() in this function, because it can interfere with destructors.
/// (Destructors are needed to ensure some piles are flushed, for example.)
fn wrapped_main() -> anyhow::Result<i32> {
    env_logger::Builder::from_env(Env::default().default_filter_or("info")).init();

    let opts: Opts = Opts::parse();

    let open_pile = || -> anyhow::Result<(PileDescriptor, Pile<Box<dyn RawPile>>)> {
        let this_dir = Path::new(".");
        let descriptor =
            load_pile_descriptor(this_dir).context("Failed to load pile descriptor")?;
        let pile = open_pile(this_dir, &descriptor).context("Failed to open pile")?;
        Ok((descriptor, pile))
    };

    match &opts.command {
        PileCommand::Retrieve {
            pointer_name,
            subset,
            destination,
            num_workers: workers,
        } => {
            let (_pdesc, pile) = open_pile()?;
            let mut pointer = pile
                .read_pointer(pointer_name)?
                .expect("No pointer by that name!"); // todo be nice

            if destination.exists() {
                bail!("The destination already exists. Overwriting not allowed (yet).");
            }

            let mut root_tree_node =
                commands::retrieve_tree_node(&pile, pointer.chunk_ref.clone())?;

            fully_integrate_pointer_node(&pile, &mut root_tree_node.node, &mut pointer)?;

            // todo allow disabling apply metadata
            extracting::extract(
                destination,
                &mut root_tree_node.node,
                &pile,
                true,
                workers.unwrap_or(2),
                true,
                true,
                true,
            )?;
        }
        PileCommand::Check {
            apply_gc,
            dry_run_gc,
            deep,
            shallow,
        } => {
            let vacuum_mode = if *dry_run_gc {
                VacuumMode::DryRunVacuum
            } else if *apply_gc {
                VacuumMode::Vacuum
            } else {
                VacuumMode::NoVacuum
            };
            let (_pdesc, pile) = open_pile()?;
            let error_count = if *deep {
                checking::check_deep(pile, vacuum_mode, true)?
            } else if *shallow {
                checking::check_shallow(pile, vacuum_mode, true, true)?
            } else {
                bail!("You need to specify either --shallow or --deep.");
            };

            if error_count > 0 {
                eprintln!("THERE ARE {} ERRORS.", error_count);
                return Ok(1);
            }
        }
        PileCommand::Init {} => {
            commands::init(".".as_ref())?;
        }

        PileCommand::Debug { supplied_command } => {
            let (pdesc, pile) = open_pile()?;
            if supplied_command.is_empty() {
                debug::debug_prompt(pdesc, pile)?;
            } else {
                let mut prefixed_command = vec![String::from("yama-debug")];
                prefixed_command.extend(supplied_command.iter().cloned());
                match DebugCommand::try_parse_from(prefixed_command) {
                    Ok(command) => {
                        if let Err(e) = debug_command(&pdesc, &pile, command) {
                            eprintln!("Failed {:?}", e);
                            pile.flush()?;
                            return Ok(2);
                        } else {
                            pile.flush()?;
                            return Ok(0);
                        }
                    }
                    Err(err) => {
                        eprintln!("Invalid {:?}", err);
                        return Ok(3);
                    }
                }
            }
        }

        PileCommand::Push {
            pointer_name,
            other_pile_path,
        } => {
            let this_pile_path = PathBuf::from(".");

            let descriptor_this = load_pile_descriptor(".".as_ref())
                .context("Failed to load this pile descriptor")?;

            let descriptor_other = load_pile_descriptor(other_pile_path)
                .context("Failed to load foreign pile descriptor.")?;

            let bypass_level = determine_bypass_level(
                &descriptor_this,
                &this_pile_path,
                &descriptor_other,
                &other_pile_path,
            )?;

            info!("Using bypass level: {:?}", bypass_level);

            let (this_pile, this_rp_bypass) =
                open_pile_with_work_bypass(&this_pile_path, &descriptor_this, bypass_level)?;
            let (other_pile, other_rp_bypass) =
                open_pile_with_work_bypass(&other_pile_path, &descriptor_other, bypass_level)?;

            // TODO flush the pile after here yet
            push_to(
                Arc::new(this_pile),
                this_rp_bypass,
                Arc::new(other_pile),
                other_rp_bypass,
                vec![pointer_name.clone()],
                true,
                32,
            )?;
        }
    }

    Ok(0)
}
