/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::cmp::min;
use std::io;
use std::io::{Cursor, Read, Write};

use anyhow::anyhow;
use fastcdc::FastCDC;

use crate::definitions::{ChunkId, RecursiveChunkRef};
use crate::pile::{Pile, RawPile};
use crate::utils::bytes_to_hexstring;

pub const SENSIBLE_THRESHOLD: usize = 1024 * 1024;

// 256 kiB
pub const FASTCDC_MIN: usize = 256 * 1024;
// 1 MiB
pub const FASTCDC_AVG: usize = 1024 * 1024;
// 8 MiB
pub const FASTCDC_MAX: usize = 8 * 1024 * 1024;

pub trait ChunkSubmissionTarget: Sync {
    fn submit(&self, chunk_id: ChunkId, chunk_data: &[u8]) -> anyhow::Result<()>;
}

impl<RP: RawPile> ChunkSubmissionTarget for Pile<RP> {
    fn submit(&self, chunk_id: ChunkId, chunk_data: &[u8]) -> anyhow::Result<()> {
        self.submit_chunk(chunk_id, chunk_data)
    }
}

impl ChunkSubmissionTarget for crossbeam_channel::Sender<(ChunkId, Vec<u8>)> {
    fn submit(&self, chunk_id: ChunkId, chunk_data: &[u8]) -> anyhow::Result<()> {
        self.send((chunk_id, chunk_data.to_vec()))
            .map_err(|_| anyhow::anyhow!("Failed to send to pipeline."))
    }
}

/// A chunker that will generate nested chunks of chunk references if there is that much data
/// to store.
/// The root RecursiveChunker is fed data bytes.
/// If it exceeds the nominated threshold, it grows a child RecursiveChunker (which may do the same).
/// When done, `finish` should be called to flush the buffers and obtain a `RecursiveChunkRef`.
pub struct RecursiveChunker<'cst, CST: ChunkSubmissionTarget> {
    /// The pile to submit chunks to.
    target: &'cst CST,
    /// Buffer of data at this level.
    buffer: Vec<u8>,
    /// The next-layer recursive chunker. This is where this chunker will submit chunk IDs to for
    /// recursive chunking.
    next_layer: Option<Box<RecursiveChunker<'cst, CST>>>,
    /// The size at which this chunker will perform recursive chunking.
    threshold: usize,
}

impl<'cst, CST: ChunkSubmissionTarget> RecursiveChunker<'cst, CST> {
    pub fn new(threshold: usize, target: &'cst CST) -> Self {
        RecursiveChunker {
            target,
            buffer: vec![],
            next_layer: None,
            threshold,
        }
    }

    /// finalise: true iff this is the last chunk (we will not reject a chunk which may have been
    ///     truncated)
    fn do_chunking(&mut self, finalise: bool) -> anyhow::Result<Vec<u8>> {
        let fastcdc = FastCDC::new(&self.buffer, FASTCDC_MIN, FASTCDC_AVG, FASTCDC_MAX);
        let mut new_chunks: Vec<u8> = Vec::new();
        let mut consumed_until: Option<usize> = None;

        for chunk in fastcdc {
            let is_final = chunk.offset + chunk.length == self.buffer.len();
            if !is_final || finalise {
                consumed_until = Some(chunk.offset + chunk.length);
                let chunk_data = &self.buffer[chunk.offset..chunk.offset + chunk.length];
                let chunk_id = calculate_chunkid(chunk_data);
                new_chunks.extend_from_slice(&chunk_id);
                self.target.submit(chunk_id, chunk_data)?;
            }
        }

        if let Some(consumed_until) = consumed_until {
            if consumed_until > 0 {
                self.buffer.drain(0..consumed_until);
            }
        }

        Ok(new_chunks)
    }

    pub fn finish(mut self) -> anyhow::Result<RecursiveChunkRef> {
        if self.next_layer.is_some() {
            // we are chunking so make this the last chunk
            let new_chunks = self.do_chunking(true)?;
            let mut subchunker = self.next_layer.unwrap();
            subchunker.write(&new_chunks)?;
            let mut rcr = subchunker.finish()?;
            // as there is a level of chunking, increase the depth
            rcr.depth += 1;
            Ok(rcr)
        } else {
            // no chunking, so depth=0 (raw) and just emit our unchunked data
            let chunk_id = calculate_chunkid(&self.buffer);
            self.target.submit(chunk_id, &self.buffer)?;
            Ok(RecursiveChunkRef { chunk_id, depth: 0 })
        }
    }

    fn write_impl(&mut self, buf: &[u8]) -> anyhow::Result<usize> {
        self.buffer.extend_from_slice(buf);

        if self.buffer.len() > self.threshold {
            if self.next_layer.is_none() {
                // start chunking
                self.next_layer = Some(Box::new(RecursiveChunker::new(
                    self.threshold,
                    self.target.clone(),
                )));
            }

            let new_chunks = self.do_chunking(false)?;

            self.next_layer.as_mut().unwrap().write(&new_chunks)?;
        }

        Ok(buf.len())
    }
}

impl<'cst, CST: ChunkSubmissionTarget> Write for RecursiveChunker<'cst, CST> {
    fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
        match self.write_impl(buf) {
            Err(err) => Err(io::Error::new(io::ErrorKind::Other, err)),
            Ok(written) => Ok(written),
        }
    }

    fn flush(&mut self) -> io::Result<()> {
        // nop is probably the most correct action here...
        Ok(())
    }
}

#[inline]
pub fn calculate_chunkid(chunk: &[u8]) -> ChunkId {
    let mut chunk_id: ChunkId = Default::default();
    blake::hash(256, &chunk, &mut chunk_id).expect("BLAKE problem");
    chunk_id
}

pub struct RecursiveUnchunker<'pile, RP: RawPile> {
    sub_reader: Box<dyn Read + 'pile>,
    buffer: Vec<u8>,
    pile: &'pile Pile<RP>,
}

impl<'pile, RP: RawPile> RecursiveUnchunker<'pile, RP> {
    /// Create a new recursive unchunker.
    /// This will automatically create sub-unchunkers as needed.
    pub fn new(pile: &'pile Pile<RP>, reference: RecursiveChunkRef) -> Self {
        if reference.depth == 0 {
            // this unchunker only needs to unchunk the one chunk
            RecursiveUnchunker {
                sub_reader: Box::new(Cursor::new(reference.chunk_id.to_vec())),
                buffer: vec![],
                pile: &pile,
            }
        } else {
            let next_ref = RecursiveChunkRef {
                depth: reference.depth - 1,
                ..reference
            };
            let sub_unchunker = RecursiveUnchunker::new(pile, next_ref);
            RecursiveUnchunker {
                sub_reader: Box::new(sub_unchunker),
                buffer: vec![],
                pile: &pile,
            }
        }
    }
}

impl<'pile, RP: RawPile> Read for RecursiveUnchunker<'pile, RP> {
    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
        while self.buffer.is_empty() {
            // Internal buffer is empty, so we need to load another chunk.
            // (we use 'while' in case we load an empty chunk...)
            let mut next_chunk_id: ChunkId = Default::default();
            let read_bytes = self.sub_reader.read(&mut next_chunk_id)?;

            if read_bytes == 0 {
                // end of chunks, because return of zero here means EOF
                return Ok(0);
            }

            if read_bytes < next_chunk_id.len() {
                // any error, including EOF at this point, is an error
                self.sub_reader
                    .read_exact(&mut next_chunk_id[read_bytes..])?;
            }

            let chunk = self
                .pile
                .read_chunk(&next_chunk_id)
                .map_err(|e| io::Error::new(io::ErrorKind::Other, e))?;
            if let Some(chunk) = chunk {
                self.buffer.extend_from_slice(&chunk);
            } else {
                return Err(io::Error::new(
                    io::ErrorKind::NotFound,
                    anyhow!("Chunk {} not found", bytes_to_hexstring(&next_chunk_id)),
                ));
            }
        }

        let to_read = min(self.buffer.len(), buf.len());
        buf[0..to_read].copy_from_slice(&self.buffer[0..to_read]);
        self.buffer.drain(0..to_read);
        Ok(to_read)
    }
}
