/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use crate::commands::{fully_integrate_pointer_node, retrieve_tree_node, store_tree_node};
use crate::definitions::{FilesystemOwnership, FilesystemPermissions, TreeNode};
use crate::pile::{Pile, PileDescriptor, RawPile};
use crate::tree::integrate_node_in_place;
use anyhow::anyhow;
use clap::Parser;
use log::info;
use rustyline::error::ReadlineError;
use rustyline::Editor;

#[derive(Parser)]
pub enum DebugCommand {
    /// List the pointers that are stored in this yama pile.
    #[clap(name = "lsp")]
    ListPointers {
        /// List details about each pointer.
        #[clap(short)]
        verbose: bool,
    },
    /// Delete a pointer from the yama pile.
    #[clap(name = "rmp")]
    DeletePointer {
        /// Name of the pointer to delete.
        name: String,
    },
    /// Reads the information on a pointer.
    #[clap(name = "infop")]
    PointerInfo {
        /// Name of the pointer to read.
        name: String,
    },
    /// Reads statistics from the Pile
    #[clap(name = "stats")]
    Statistics {},
}

pub fn debug_prompt<RP: RawPile>(pdesc: PileDescriptor, pile: Pile<RP>) -> anyhow::Result<()> {
    let mut rl = Editor::<()>::new();
    if rl.load_history(".yama-history").is_err() {
        // no previous history...
    }
    loop {
        let readline = rl.readline("debug 山 ");
        match readline {
            Ok(line) => {
                rl.add_history_entry(line.as_str());
                let mut args = vec!["山"];
                args.extend(line.split_ascii_whitespace());
                match DebugCommand::try_parse_from(args) {
                    Ok(command) => {
                        if let Err(e) = debug_command(&pdesc, &pile, command) {
                            eprintln!("Failed {:?}", e);
                        }
                    }
                    Err(e) => {
                        eprintln!("{}", e);
                    }
                }
            }
            Err(ReadlineError::Interrupted) => {
                eprintln!("^C");
                break;
            }
            Err(ReadlineError::Eof) => {
                eprintln!("^D");
                break;
            }
            Err(err) => {
                eprintln!("Error: {:?}", err);
                break;
            }
        }
    }

    pile.flush()?;

    rl.save_history(".yama-history").unwrap();

    Ok(())
}

pub fn debug_command<RP: RawPile>(
    _pdesc: &PileDescriptor,
    pile: &Pile<RP>,
    command: DebugCommand,
) -> anyhow::Result<()> {
    match &command {
        DebugCommand::ListPointers { verbose } => {
            for pointer in pile.list_pointers()?.iter() {
                if *verbose {
                    let pointer_data = pile.read_pointer(pointer.as_str())?;
                    match pointer_data {
                        None => {
                            println!("{} → ??? MISSING DATA", pointer);
                        }
                        Some(data) => {
                            println!(
                                "{} → {:?} [parent={:?}]",
                                pointer, data.chunk_ref, data.parent_pointer
                            );
                        }
                    }
                } else {
                    println!("{}", pointer);
                }
            }
        }
        DebugCommand::DeletePointer { name } => {
            // retrieve this pointer
            let mut this_pointer = pile.read_pointer(name.as_str())?.ok_or_else(|| {
                anyhow!("Pointer {:?} does not exist so can not be deleted.", name)
            })?;
            let mut this_node = retrieve_tree_node(&pile, this_pointer.chunk_ref.clone())?;

            // fully integrate the pointer
            fully_integrate_pointer_node(&pile, &mut this_node.node, &mut this_pointer)?;
            assert!(this_pointer.parent_pointer.is_none());

            // now integrate any pointers that rely on this one
            // so that they no longer rely on this one.
            for pointer in pile.list_pointers()?.iter() {
                if pointer == name {
                    continue;
                }
                if let Some(mut pointer_data) = pile.read_pointer(pointer.as_str())? {
                    if let Some(parent_pointer) = pointer_data.parent_pointer.as_ref() {
                        if parent_pointer == name {
                            info!("Pointer is now an orphan: {:?}", pointer);

                            // need to integrate this node, so retrieve it
                            let mut node = retrieve_tree_node(&pile, pointer_data.chunk_ref)?;

                            // integrate it in-place
                            integrate_node_in_place(&mut node.node, &this_node.node)?;

                            // mark it as orphaned (no parent)
                            pointer_data.parent_pointer = None;

                            // store the orphaned node
                            let new_chunk_ref = store_tree_node(&pile, &node)?;
                            // associate the orphaned node with the orphaned pointer
                            pointer_data.chunk_ref = new_chunk_ref;
                            // write the pointer back.
                            pile.write_pointer(pointer.as_str(), &pointer_data)?;
                        }
                    }
                }
            }

            // then delete the pointer
            pile.delete_pointer(name)?;
            info!("Deleted pointer: {:?}", name);
        }
        DebugCommand::PointerInfo { name } => {
            let this_pointer = pile
                .read_pointer(name.as_str())?
                .ok_or_else(|| anyhow!("Pointer {:?} does not exist.", name))?;
            let this_node = retrieve_tree_node(&pile, this_pointer.chunk_ref.clone())?;
            eprintln!(" --- Pointer data --- ");
            eprintln!("{:#?}", this_pointer);
            eprintln!(" --- Tree node --- ");
            //eprintln!("{:#?}", this_node.node);
            tree_node_printer(&this_node.name, &this_node.node, 0);
        }
        DebugCommand::Statistics { .. } => {
            if let Some(stats) = pile.raw_pile.debug_statistics()? {
                println!("Statistics for this pile");
                println!("  chunk count: {} chunks", stats.number_of_chunks);
                println!(
                    "  total chunk stored space: {} bytes (may exclude deleted chunks)",
                    stats.total_chunk_size
                );
                let average_size =
                    ((stats.total_chunk_size as f64) / (stats.number_of_chunks as f64)) as u64;
                println!("    (average chunk size: {} bytes)", average_size);
            } else {
                eprintln!("{:?}", pile);
                eprintln!("Statistics appear not to be supported on this kind of pile?");
            }
        }
    }
    Ok(())
}

pub fn compact_ownership(ownership: &FilesystemOwnership) -> String {
    format!("uid={}, gid={}", ownership.uid, ownership.gid)
}

pub fn compact_permissions(perms: &FilesystemPermissions) -> String {
    format!("{:4o}", perms.mode)
}

pub fn tree_node_printer(name: &str, node: &TreeNode, depth: usize) {
    match node {
        TreeNode::NormalFile {
            mtime,
            ownership,
            permissions,
            content,
        } => {
            eprintln!(
                "{}{}       = {:?} ({}, {}, mtime={})",
                "  ".repeat(depth),
                name,
                content,
                compact_ownership(ownership),
                compact_permissions(permissions),
                mtime
            );
        }
        TreeNode::Directory {
            ownership,
            permissions,
            children,
        } => {
            eprintln!(
                "{}{}/       ({}, {})",
                "  ".repeat(depth),
                name,
                compact_ownership(ownership),
                compact_permissions(permissions)
            );
            for (name, child) in children.iter() {
                tree_node_printer(name, child, depth + 1);
            }
        }
        TreeNode::SymbolicLink { ownership, target } => {
            eprintln!(
                "{}{} → {}       ({})",
                "  ".repeat(depth),
                name,
                target,
                compact_ownership(ownership)
            );
        }
        TreeNode::Deleted => {
            eprintln!("{}{}       DELETED", "  ".repeat(depth), name);
        }
    }
}
