/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::collections::BTreeMap;
use std::fmt::{Debug, Formatter};

use anyhow::bail;
use serde::{Deserialize, Serialize};

use crate::utils::bytes_to_hexstring;

pub type ChunkId = [u8; 32];
pub type XXHash = u64;
pub const XXH64_SEED: u64 = 424242;

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct PointerData {
    pub chunk_ref: RecursiveChunkRef,
    pub parent_pointer: Option<String>,
    pub uid_lookup: BTreeMap<u16, Option<String>>,
    pub gid_lookup: BTreeMap<u16, Option<String>>,
}

#[derive(Clone, Serialize, Deserialize, PartialEq, Eq, Hash)]
pub struct RecursiveChunkRef {
    /// The root Chunk ID.
    pub chunk_id: ChunkId,
    /// The depth of the data bytes.
    /// 0 means that the chunk addressed by `chunk_id` contains data bytes.
    /// 1 means that the chunk addressed by `chunk_id` contains references to chunk that contain
    /// data bytes.
    /// (and so on)
    pub depth: u32,
}

impl Debug for RecursiveChunkRef {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}<{}>", bytes_to_hexstring(&self.chunk_id), self.depth)
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct RootTreeNode {
    pub name: String,
    pub node: TreeNode,
}

#[derive(Debug, Clone, Serialize, Deserialize, Eq, PartialEq)]
pub enum TreeNode {
    NormalFile {
        /// modification time in ms
        mtime: u64,
        ownership: FilesystemOwnership,
        permissions: FilesystemPermissions,
        // TODO size: u64 or not
        // can perhaps cache chunk-wise (but not sure.)
        content: RecursiveChunkRef,
    },
    Directory {
        ownership: FilesystemOwnership,
        permissions: FilesystemPermissions,
        children: BTreeMap<String, TreeNode>,
    },
    SymbolicLink {
        ownership: FilesystemOwnership,
        target: String,
    },
    // TODO is there any other kind of file we need to store?
    Deleted,
}

impl TreeNode {
    pub fn metadata_invalidates(&self, other: &TreeNode) -> anyhow::Result<bool> {
        Ok(match self {
            TreeNode::NormalFile {
                mtime,
                ownership,
                permissions,
                ..
            } => {
                if let TreeNode::NormalFile {
                    mtime: other_mtime,
                    ownership: other_ownership,
                    permissions: other_permissions,
                    ..
                } = other
                {
                    mtime != other_mtime
                        || ownership != other_ownership
                        || permissions != other_permissions
                } else {
                    true
                }
            }
            TreeNode::Directory {
                ownership,
                permissions,
                children,
            } => {
                if let TreeNode::Directory {
                    ownership: other_ownership,
                    permissions: other_permissions,
                    children: other_children,
                } = other
                {
                    if ownership != other_ownership || permissions != other_permissions {
                        return Ok(true);
                    }
                    if children.len() != other_children.len() {
                        return Ok(true);
                    }
                    for ((left_name, left_node), (right_name, right_node)) in
                        children.iter().zip(other_children.iter())
                    {
                        if left_name != right_name || left_node.metadata_invalidates(right_node)? {
                            return Ok(true);
                        }
                    }
                    false
                } else {
                    true
                }
            }
            TreeNode::SymbolicLink { ownership, target } => {
                if let TreeNode::SymbolicLink {
                    ownership: other_ownership,
                    target: other_target,
                } = other
                {
                    ownership != other_ownership || target != other_target
                } else {
                    true
                }
            }
            TreeNode::Deleted => {
                // unreachable
                bail!("Why is Deleted here?");
            }
        })
    }
    //
    ///// Guarantees consistent visit order.
    // pub fn visit_mut<F>(
    //     &mut self,
    //     visitor: &mut F,
    //     path_prefix: &str,
    //     skip_components: u32,
    // ) -> anyhow::Result<()>
    //     where
    //         F: FnMut(&mut Self, &str) -> anyhow::Result<()>,
    // {
    //     let mut my_path_buf = String::new();
    //     my_path_buf.push_str(path_prefix);
    //
    //     if skip_components == 0 {
    //         if !my_path_buf.is_empty() {
    //             my_path_buf.push('/');
    //         }
    //         my_path_buf.push_str(&self.name);
    //     }
    //
    //     visitor(self, &my_path_buf)?;
    //
    //     if let TreeNode::Directory { children, .. } = &mut self.content {
    //         for child in children.iter_mut() {
    //             let new_skip = if skip_components > 0 {
    //                 skip_components - 1
    //             } else {
    //                 0
    //             };
    //             child.visit_mut(visitor, &my_path_buf, new_skip)?;
    //         }
    //     }
    //     Ok(())
    // }

    pub fn count_normal_files(&self) -> u32 {
        match self {
            TreeNode::NormalFile { .. } => 1,
            TreeNode::Directory { children, .. } => children
                .values()
                .map(|child| child.count_normal_files())
                .sum(),
            _ => 0,
        }
    }

    pub fn visit<F: FnMut(&TreeNode, &str) -> anyhow::Result<()>>(
        &self,
        func: &mut F,
        prefix: String,
    ) -> anyhow::Result<()> {
        func(self, &prefix)?;
        if let TreeNode::Directory { children, .. } = &self {
            for (name, child) in children.iter() {
                if prefix.is_empty() {
                    // don't want a slash prefix
                    child.visit(func, name.clone())?;
                } else {
                    child.visit(func, format!("{}/{}", prefix, name))?;
                }
            }
        }
        Ok(())
    }

    pub fn visit_mut<F: FnMut(&mut TreeNode, &str) -> anyhow::Result<()>>(
        &mut self,
        func: &mut F,
        prefix: String,
    ) -> anyhow::Result<()> {
        func(self, &prefix)?;
        if let TreeNode::Directory { children, .. } = self {
            for (name, child) in children.iter_mut() {
                if prefix.is_empty() {
                    // don't want a slash prefix
                    child.visit_mut(func, name.clone())?;
                } else {
                    child.visit_mut(func, format!("{}/{}", prefix, name))?;
                }
            }
        }
        Ok(())
    }

    pub fn exists(&self, pieces: &[&str]) -> bool {
        match pieces.first() {
            None => true,
            Some(subpath) => {
                if let TreeNode::Directory { children, .. } = self {
                    match children.get(*subpath) {
                        None => false,
                        Some(child) => child.exists(&pieces[1..]),
                    }
                } else {
                    false
                }
            }
        }
    }
}

#[derive(Debug, Copy, Clone, Serialize, Deserialize, PartialEq, Eq)]
pub struct FilesystemOwnership {
    pub uid: u16,
    pub gid: u16,
}

#[derive(Debug, Copy, Clone, Serialize, Deserialize, PartialEq, Eq)]
pub struct FilesystemPermissions {
    pub mode: u32,
}

#[cfg(test)]
pub mod tests {
    use crate::definitions::{
        FilesystemOwnership, FilesystemPermissions, RecursiveChunkRef, TreeNode,
    };
    use std::collections::BTreeMap;

    pub fn example_file() -> TreeNode {
        TreeNode::NormalFile {
            mtime: 424242,
            ownership: FilesystemOwnership {
                uid: 1042,
                gid: 1043,
            },
            permissions: FilesystemPermissions { mode: 0o760 },
            content: RecursiveChunkRef {
                chunk_id: Default::default(),
                depth: 0,
            },
        }
    }

    pub fn example_dir(
        file1: Option<(&str, TreeNode)>,
        file2: Option<(&str, TreeNode)>,
    ) -> TreeNode {
        let mut map = BTreeMap::new();
        if let Some((name, file)) = file1 {
            map.insert(name.to_owned(), file);
        }
        if let Some((name, file)) = file2 {
            map.insert(name.to_owned(), file);
        }
        TreeNode::Directory {
            ownership: FilesystemOwnership {
                uid: 1042,
                gid: 1043,
            },
            permissions: FilesystemPermissions { mode: 0o770 },
            children: map,
        }
    }

    #[test]
    pub fn test_exists() {
        let file = example_file();
        assert!(file.exists(&[]));
        assert!(!file.exists(&["anything"]));

        let subdir = example_dir(Some(("fetchmailrc", example_file())), None);
        let dir = example_dir(Some(("boot.img", example_file())), Some(("etc", subdir)));
        assert!(dir.exists(&[]));
        assert!(dir.exists(&["boot.img"]));
        assert!(dir.exists(&["etc", "fetchmailrc"]));
        assert!(!dir.exists(&["bin"]));
        assert!(!dir.exists(&["etc", "resolv.conf"]));
        assert!(!dir.exists(&["boot.img", "hehe"]));
    }
}
