/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::convert::TryInto;
use std::fs::OpenOptions;
use std::os::unix::fs::PermissionsExt;
use std::path::{Path, PathBuf};
use std::sync::atomic::{AtomicU32, Ordering};
use std::{fs, io};

use anyhow::{anyhow, Context};
use crossbeam_channel::{Receiver, Sender};
use crossbeam_utils::thread;
use indicatif::{ProgressBar, ProgressDrawTarget, ProgressStyle};
use log::error;
use nix::sys::time::{TimeVal, TimeValLike};
use nix::unistd::{Gid, Uid};

use crate::chunking::RecursiveUnchunker;
use crate::commands::fully_load_pointer;
use crate::definitions::{FilesystemOwnership, RecursiveChunkRef, TreeNode};
use crate::pile::{Pile, RawPile};
use std::collections::{BTreeMap, HashMap};

/// Given a fully-integrated root node, extracts the files from the pile.
pub fn extract<RP: RawPile>(
    target_path: &Path,
    root: &mut TreeNode,
    pile: &Pile<RP>,
    make_progress_bar: bool,
    num_workers: u8,
    apply_permissions: bool,
    apply_mtime: bool,
    apply_ownership: bool,
) -> anyhow::Result<()> {
    let pbar = if make_progress_bar {
        ProgressBar::with_draw_target(
            root.count_normal_files() as u64,
            ProgressDrawTarget::stdout_with_hz(10),
        )
    } else {
        ProgressBar::hidden()
    };
    pbar.set_style(
        ProgressStyle::default_bar()
            .template("[{elapsed_precise}]/[{eta}] {bar:40.cyan/blue} {pos:>7}/{len:7} {msg}"),
    );
    pbar.set_message("extracting");

    let (paths_send, paths_recv) = crossbeam_channel::unbounded();
    let (results_send, results_recv) = crossbeam_channel::bounded(16);

    let failures = AtomicU32::new(0);

    thread::scope(|s| {
        for worker in 0..num_workers {
            let paths_recv = paths_recv.clone();
            let results_send = results_send.clone();
            let failures = &failures; // needed because of move
            s.builder()
                .name(format!("yama unchunker {}", worker))
                .spawn(move |_| {
                    if let Err(e) = extract_worker(pile, paths_recv, results_send) {
                        error!("Extraction worker {} failed: {:?}!", worker, e);
                        failures.fetch_add(1, Ordering::Relaxed);
                    }
                })
                .expect("Failed to start thread");
        }

        // Needed to allow the manager to join once the workers finish and drop their senders.
        drop(results_send);
        drop(paths_recv);

        s.spawn(|_| {
            if let Err(e) = manager(root, target_path, paths_send, results_recv, &pbar) {
                error!("Extraction manager failed: {:?}!", e);
                failures.fetch_add(1, Ordering::Relaxed);
            }
        });
    })
    .expect("join issue");

    pbar.set_message("applying metadata");

    apply_metadata(
        root,
        target_path,
        apply_permissions,
        apply_mtime,
        apply_ownership,
    )?;

    Ok(())
}

/// Given the name of a pointer, extracts it.
pub fn extract_from_pointer_name<RP: RawPile>(
    target_path: &Path,
    pointer_name: &str,
    pile: &Pile<RP>,
    make_progress_bar: bool,
    num_workers: u8,
    apply_permissions: bool,
    apply_mtime: bool,
    apply_ownership: bool,
) -> anyhow::Result<()> {
    let (pointer_data, mut root_node) = fully_load_pointer(pile, pointer_name.as_ref())?;

    let uid_translation_table = build_uid_translation_table(&pointer_data.uid_lookup);
    let gid_translation_table = build_gid_translation_table(&pointer_data.gid_lookup);

    // convert the UIDs and GIDs to match this system, which may be different from the usual.
    apply_uid_and_gid_translation_tables(
        &mut root_node.node,
        &uid_translation_table,
        &gid_translation_table,
    );

    extract(
        &target_path.join(&root_node.name),
        &mut root_node.node,
        pile,
        make_progress_bar,
        num_workers,
        apply_permissions,
        apply_mtime,
        apply_ownership,
    )
}

pub fn build_uid_translation_table(
    uid_lookup: &BTreeMap<u16, Option<String>>,
) -> HashMap<u16, u16> {
    let mut result: HashMap<u16, u16> = Default::default();

    for (old_uid, name) in uid_lookup.iter() {
        if let Some(name) = name {
            if let Some(user) = users::get_user_by_name(name) {
                let new_uid = user.uid() as u16;
                if new_uid != *old_uid {
                    result.insert(*old_uid, new_uid);
                }
            }
        }
    }

    result
}

pub fn build_gid_translation_table(
    gid_lookup: &BTreeMap<u16, Option<String>>,
) -> HashMap<u16, u16> {
    let mut result: HashMap<u16, u16> = Default::default();

    for (old_gid, name) in gid_lookup.iter() {
        if let Some(name) = name {
            if let Some(group) = users::get_group_by_name(name) {
                let new_gid = group.gid() as u16;
                if new_gid != *old_gid {
                    result.insert(*old_gid, new_gid);
                }
            }
        }
    }

    result
}

pub fn apply_uid_and_gid_translation_tables(
    node: &mut TreeNode,
    uid_translation: &HashMap<u16, u16>,
    gid_translation: &HashMap<u16, u16>,
) {
    if uid_translation.is_empty() && gid_translation.is_empty() {
        // nothing to do here :).
        return;
    }

    let apply_to = |ownership: &mut FilesystemOwnership| {
        ownership.uid = *uid_translation
            .get(&ownership.uid)
            .unwrap_or(&ownership.uid);
        ownership.gid = *gid_translation
            .get(&ownership.gid)
            .unwrap_or(&ownership.gid);
    };
    node.visit_mut(
        &mut |node, _| {
            match node {
                TreeNode::NormalFile { ownership, .. } => {
                    apply_to(ownership);
                }
                TreeNode::Directory { ownership, .. } => {
                    apply_to(ownership);
                }
                TreeNode::SymbolicLink { ownership, .. } => {
                    apply_to(ownership);
                }
                TreeNode::Deleted => {}
            }
            Ok(())
        },
        "".to_owned(),
    )
    .expect("Can't fail since we don't fail.");
}

/// A worker thread for extracting
pub fn extract_worker<RP: RawPile>(
    pile: &Pile<RP>,
    paths: Receiver<(PathBuf, RecursiveChunkRef)>,
    results: Sender<()>,
) -> anyhow::Result<()> {
    while let Ok((path, chunk_ref)) = paths.recv() {
        let mut extractor = RecursiveUnchunker::new(pile, chunk_ref);
        let mut file = OpenOptions::new()
            .write(true)
            .create_new(true)
            .open(&path)
            .with_context(|| format!("Failed to open {:?}", path))?;

        io::copy(&mut extractor, &mut file)?;

        results
            .send(())
            .or_else(|_| Err(anyhow!("Failed to send result")))?;
    }
    Ok(())
}

/// A single thread that manages the workers
pub fn manager(
    root: &mut TreeNode,
    target_path: &Path,
    paths_sender: Sender<(PathBuf, RecursiveChunkRef)>,
    results_receiver: Receiver<()>,
    progress_bar: &ProgressBar,
) -> anyhow::Result<()> {
    root.visit(
        &mut |tree_node, name| {
            let final_path = if name.is_empty() {
                target_path.to_path_buf()
            } else {
                target_path.join(name)
            };
            match tree_node {
                TreeNode::NormalFile { content, .. } => {
                    paths_sender
                        .send((final_path, content.clone()))
                        .expect("Unable to send to should-be unbounded channel");
                }
                TreeNode::Directory { .. } => {
                    fs::create_dir(&final_path)?;
                }
                TreeNode::SymbolicLink { target, .. } => {
                    // TODO may want to perform rewrites ...?
                    std::os::unix::fs::symlink(target, &final_path)?;
                }
                TreeNode::Deleted => {
                    panic!("should not be extracting 'Deleted!' --- BUG.");
                }
            };
            Ok(())
        },
        "".to_string(),
    )?;

    // Needed to allow the workers to finish; otherwise we never join.
    drop(paths_sender);

    while let Ok(()) = results_receiver.recv() {
        progress_bar.inc(1);
    }

    Ok(())
}

/// Applies metadata (permissions, mtime, ownership) to files from a tree node.
pub fn apply_metadata(
    root: &TreeNode,
    target: &Path,
    apply_permissions: bool,
    apply_mtime: bool,
    apply_owner: bool,
) -> anyhow::Result<()> {
    match root {
        TreeNode::NormalFile {
            mtime,
            ownership,
            permissions,
            ..
        } => {
            if apply_permissions {
                let mut perms = fs::metadata(&target)?.permissions();
                perms.set_mode(permissions.mode);
                fs::set_permissions(&target, perms)?;
            }
            if apply_owner {
                nix::unistd::chown(
                    target,
                    Some(Uid::from_raw(ownership.uid as u32)),
                    Some(Gid::from_raw(ownership.gid as u32)),
                )?;
            }
            if apply_mtime {
                if let Ok(mtime) = (*mtime).try_into() {
                    let tv = TimeVal::milliseconds(mtime);
                    nix::sys::stat::lutimes(target, &tv, &tv)?;
                }
            }
        }
        TreeNode::Directory {
            ownership,
            permissions,
            children,
        } => {
            if apply_permissions {
                let mut perms = fs::metadata(&target)?.permissions();
                perms.set_mode(permissions.mode);
                fs::set_permissions(&target, perms)?;
            }
            if apply_owner {
                nix::unistd::chown(
                    target,
                    Some(Uid::from_raw(ownership.uid as u32)),
                    Some(Gid::from_raw(ownership.gid as u32)),
                )?;
            }
            for (name, child) in children.iter() {
                let child_path = target.join(name);
                apply_metadata(
                    child,
                    &child_path,
                    apply_permissions,
                    apply_mtime,
                    apply_owner,
                )?;
            }
        }
        TreeNode::SymbolicLink { ownership, .. } => {
            if apply_owner {
                nix::unistd::chown(
                    target,
                    Some(Uid::from_raw(ownership.uid as u32)),
                    Some(Gid::from_raw(ownership.gid as u32)),
                )?;
            }
        }
        TreeNode::Deleted => {
            panic!("Deleted is not meant to be reachable here.");
        }
    }

    Ok(())
}
