use crate::chunking::RecursiveUnchunker;
use crate::commands::fully_load_pointer;
use crate::definitions::{ChunkId, RecursiveChunkRef, TreeNode};
use crate::operations::checking::VacuumRawPile;
use crate::operations::pushpull::PushWorkerToManagerMessage::{NewTask, TaskDone};
use crate::pile::compression::{CompressionSettings, RawPileCompressor};
use crate::pile::integrity::RawPileIntegrityChecker;
use crate::pile::local_sqlitebloblogs::SqliteBloblogPile;
use crate::pile::{Keyspace, Pile, PileDescriptor, PileStorage, RawPile};
use crate::utils::get_number_of_workers;
use anyhow::{bail, Context};
use crossbeam_channel::{Receiver, Sender};
use indicatif::{ProgressBar, ProgressDrawTarget, ProgressStyle};
use log::error;
use std::collections::HashSet;
use std::fs::File;
use std::io::Read;
use std::path::Path;
use std::sync::atomic::{AtomicU32, Ordering};
use std::sync::Arc;

/// Pushes chunks (and pointers) from one pile to another.
/// This is a thorough implementation that could be slow but at least should give good confidence.
/// (Presumably we could do better by looking at the pointers that already exist on the destination
///  and only integrating as much as we need to.)
pub fn push_to(
    from_pile: Arc<Pile<Arc<Box<dyn RawPile>>>>,
    from_rp_bypass: Arc<Box<dyn RawPile>>,
    to_pile: Arc<Pile<Arc<Box<dyn RawPile>>>>,
    to_rp_bypass: Arc<Box<dyn RawPile>>,
    pointers: Vec<String>,
    make_progress_bar: bool,
    num_workers: u32,
) -> anyhow::Result<()> {
    let pbar = if make_progress_bar {
        ProgressBar::with_draw_target(
            1, // TODO
            ProgressDrawTarget::stdout_with_hz(10),
        )
    } else {
        ProgressBar::hidden()
    };
    pbar.set_style(
        ProgressStyle::default_bar()
            .template("[{elapsed_precise}]/[{eta}] {bar:40.cyan/blue} {pos:>7}/{len:7} {msg}"),
    );
    pbar.set_message("push/pull");

    let (jobs_tx, jobs_rx) = crossbeam_channel::unbounded();
    let (stat_tx, stat_rx) = crossbeam_channel::bounded(32);

    let mut to_process = Vec::new();

    for pointer in pointers {
        let (pointer_data, root_node) = fully_load_pointer(&from_pile, &pointer)?;

        // schedule storing the pointer chunks
        to_process.push(pointer_data.chunk_ref.clone());

        if to_pile.read_pointer(&pointer)?.is_some() {
            unimplemented!("pointer in target exists.");
        }

        // copy across the pointer data
        to_pile.write_pointer(&pointer, &pointer_data)?;

        root_node
            .node
            .visit(
                &mut |node, _path| {
                    match node {
                        TreeNode::NormalFile { content, .. } => {
                            to_process.push(content.clone());
                        }
                        _ => {} // nop
                    }
                    Ok(())
                },
                String::new(),
            )
            .expect("No fail");
    }

    // start the work

    let critical_failures = Arc::new(AtomicU32::new(0));

    for worker_num in 0..num_workers {
        let jobs_rx = jobs_rx.clone();
        let stat_tx = stat_tx.clone();
        let critical_failures = critical_failures.clone();

        let from_pile = from_pile.clone();
        let from_rp_bypass = from_rp_bypass.clone();
        let to_pile = to_pile.clone();
        let to_rp_bypass = to_rp_bypass.clone();

        std::thread::Builder::new()
            .name(format!("yama pusher {}", worker_num))
            .spawn(move || {
                if let Err(e) = pusher_worker(
                    from_pile,
                    from_rp_bypass,
                    to_pile,
                    to_rp_bypass,
                    jobs_rx,
                    stat_tx,
                ) {
                    error!("[critical!] Push worker {} FAILED: {:?}", worker_num, e);
                    critical_failures.fetch_add(1, Ordering::Relaxed);
                }
            })
            .expect("Failed to start thread");
    }

    for task in to_process {
        stat_tx
            .send(NewTask(task))
            .expect("unbounded so should be able to send");
    }

    // must drop here for ending to happen
    drop(jobs_rx);
    drop(stat_tx);

    pbar.set_length(0);
    if let Err(e) = pusher_manager(&pbar, stat_rx, jobs_tx) {
        error!("[critical!] Push manager FAILED: {:?}", e);
        critical_failures.fetch_add(1, Ordering::Relaxed);
    }

    Ok(())
}

enum PushWorkerToManagerMessage {
    NewTask(RecursiveChunkRef),
    TaskDone,
}

fn pusher_manager(
    pbar: &ProgressBar,
    update_receiver: Receiver<PushWorkerToManagerMessage>,
    job_queue: Sender<RecursiveChunkRef>,
) -> anyhow::Result<()> {
    let mut outstanding = 0;
    let mut already_done = HashSet::new();
    while let Ok(status) = update_receiver.recv() {
        match status {
            PushWorkerToManagerMessage::NewTask(task) => {
                if already_done.insert(task.clone()) {
                    job_queue.send(task)?;
                    pbar.inc_length(1);
                    outstanding += 1;
                }
            }
            PushWorkerToManagerMessage::TaskDone => {
                pbar.inc(1);
                outstanding -= 1;
                if outstanding == 0 {
                    break;
                }
            }
        }
    }
    Ok(())
}

fn pusher_worker(
    from_pile: Arc<Pile<Arc<Box<dyn RawPile>>>>,
    from_rp_bypass: Arc<Box<dyn RawPile>>,
    to_pile: Arc<Pile<Arc<Box<dyn RawPile>>>>,
    to_rp_bypass: Arc<Box<dyn RawPile>>,
    jobs_rx: Receiver<RecursiveChunkRef>,
    stat_tx: Sender<PushWorkerToManagerMessage>,
) -> anyhow::Result<()> {
    while let Ok(job) = jobs_rx.recv() {
        if !to_pile.chunk_exists(&job.chunk_id)? {
            if let Some(bypass_chunk_data) = from_rp_bypass.read(Keyspace::Chunk, &job.chunk_id)? {
                to_rp_bypass.write(Keyspace::Chunk, &job.chunk_id, &bypass_chunk_data)?;
            } else {
                bail!("Chunk cannot be copied because doesn't exist (in bypass pile).");
            }
        }

        if job.depth > 0 {
            // we want to (partially) unchunk this and submit all subchunks.

            let vacuum_rp = VacuumRawPile::new(from_pile.raw_pile.clone(), true);
            let vacuum_pile = Pile::new(vacuum_rp);

            // First read the bottom-level chunk IDs
            let mut reduced_height = job.clone();
            reduced_height.depth -= 1;

            let mut chunk_id_buf: ChunkId = Default::default();

            let mut unchunker = RecursiveUnchunker::new(&vacuum_pile, reduced_height);
            loop {
                let read_bytes = unchunker.read(&mut chunk_id_buf)?;

                if read_bytes == 0 {
                    // end of chunks, because return of zero here means EOF
                    break;
                }

                if read_bytes < chunk_id_buf.len() {
                    // any error, including EOF at this point, is an error
                    unchunker.read_exact(&mut chunk_id_buf[read_bytes..])?;
                }

                stat_tx
                    .send(NewTask(RecursiveChunkRef {
                        chunk_id: chunk_id_buf.clone(),
                        depth: 0,
                    }))
                    .expect("Should be able to send");
            }

            // Then track the chunks that we read whilst doing the above
            for needed_chunk_id in vacuum_pile
                .raw_pile
                .retrieved_chunks
                .lock()
                .expect("Should be able to lock")
                .iter()
            {
                if needed_chunk_id != &job.chunk_id {
                    // only track them if they're not the same as the one on this job.
                    stat_tx
                        .send(NewTask(RecursiveChunkRef {
                            chunk_id: needed_chunk_id.clone(),
                            depth: 0,
                        }))
                        .expect("Should be able to send");
                }
            }
        }

        stat_tx.send(TaskDone)?;
    }
    Ok(())
}

#[derive(Copy, Clone, Debug)]
pub enum BypassLevel {
    NoBypass,
    CompressionBypass,
}

pub fn determine_bypass_level(
    desc1: &PileDescriptor,
    dir1: &Path,
    desc2: &PileDescriptor,
    dir2: &Path,
) -> anyhow::Result<BypassLevel> {
    if desc1.compression.is_some() && desc2.compression.is_some() {
        let mut dictionary1 = Vec::new();
        let dict_path1 = dir1.join("important_zstd.dict");
        File::open(dict_path1)
            .context("You need important_zstd.dict in your pile folder.")?
            .read_to_end(&mut dictionary1)?;

        let mut dictionary2 = Vec::new();
        let dict_path2 = dir2.join("important_zstd.dict");
        File::open(dict_path2)
            .context("You need important_zstd.dict in your pile folder.")?
            .read_to_end(&mut dictionary2)?;

        if dictionary1 == dictionary2 {
            // we can only bypass if both dictionaries are the same
            Ok(BypassLevel::CompressionBypass)
        } else {
            Ok(BypassLevel::NoBypass)
        }
    } else {
        Ok(BypassLevel::NoBypass)
    }
}

/// Opens a pile with potential for returning a 'complete' pile as well as a lower-level 'bypass'
/// pile, which, for example, skips performing compression operations.
///
/// Return tuple: (actual pile, bypass raw pile)
pub fn open_pile_with_work_bypass(
    dir: &Path,
    desc: &PileDescriptor,
    bypass_level: BypassLevel,
) -> anyhow::Result<(Pile<Arc<Box<dyn RawPile>>>, Arc<Box<dyn RawPile>>)> {
    let num_compressors = get_number_of_workers("YAMA_COMPRESSORS");
    let num_decompressors = get_number_of_workers("YAMA_DECOMPRESSORS");

    match desc.storage {
        PileStorage::RemoteOnly => {
            bail!("This is a remote-only pile. No local storage allowed.");
        }
        PileStorage::SqliteIndexedBloblog => {
            let blob_raw_pile = RawPileIntegrityChecker::new(SqliteBloblogPile::open(dir)?);

            match bypass_level {
                BypassLevel::NoBypass => {
                    unimplemented!()
                }
                BypassLevel::CompressionBypass => {
                    let common_raw_pile: Arc<Box<dyn RawPile>> = Arc::new(Box::new(blob_raw_pile));

                    let raw_pile: Arc<Box<dyn RawPile>> = match desc.compression {
                        None => common_raw_pile.clone(),
                        Some(comp_level) => {
                            let mut dictionary = Vec::new();
                            let dict_path = dir.join("important_zstd.dict");
                            File::open(dict_path)
                                .context("You need important_zstd.dict in your pile folder.")?
                                .read_to_end(&mut dictionary)?;

                            let (compressed_pile, _handles) = RawPileCompressor::new(
                                common_raw_pile.clone(),
                                CompressionSettings {
                                    dictionary: Arc::new(dictionary),
                                    level: comp_level as i32,
                                    num_compressors: num_compressors as u32,
                                    num_decompressors: num_decompressors as u32,
                                },
                            )?;

                            Arc::new(Box::new(compressed_pile))
                        }
                    };
                    Ok((Pile::new(raw_pile), common_raw_pile))
                }
            }
        }
    }
}
