/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::path::PathBuf;

use serde::{Deserialize, Serialize};

use crate::definitions::{ChunkId, PointerData};
use crate::utils::get_number_of_workers;
use crossbeam_channel::Sender;
use std::collections::HashSet;
use std::fmt::Debug;
use std::sync::{Arc, Condvar, Mutex};

pub mod compression;
pub mod encryption;
pub mod integrity;
pub mod local_sqlitebloblogs;

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct PileDescriptor {
    /// The last version of yama that was used with this pile.
    pub yama_version: String,
    /// The storage backend to use.
    pub storage: PileStorage,
    /// If specified, the compression level of the pile.
    pub compression: Option<u16>,
}

#[derive(Serialize, Deserialize, Debug, Copy, Clone)]
pub enum PileStorage {
    /// No local storage. Pile is only usable for remotes.
    RemoteOnly,
    /// Local storage backed by bloblogs that are indexed by a SQLite database.
    SqliteIndexedBloblog,
    // Local temporary storage in which chunks are only kept for long enough to send them to
    // remotes. Unimplemented at present.
    // TODO THIS IS NOT THE CORRECT NAME ANYWAY BarePushSled,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct RemoteDescriptor {
    pub encrypted: bool,
    pub host: Option<String>,
    pub user: Option<String>,
    pub path: PathBuf,
}

#[derive(PartialOrd, PartialEq, Copy, Clone, Serialize, Deserialize, Eq)]
pub enum Keyspace {
    Chunk,
    ChunkHash,
    Pointer,
}

/// Useful information for humans. Doesn't need to be spot on, but kind of interesting.
#[derive(Debug, Clone)]
pub struct DebugStatistics {
    pub number_of_chunks: u64,
    pub minimum_chunk_size: Option<u32>,
    pub maximum_chunk_size: Option<u32>,
    pub total_chunk_size: u64,
}

#[derive(Debug, Clone)]
pub struct StoragePipelineSettings {
    pub num_compressors: u32,
    pub compressor_input_bound: u32,
    pub writer_input_bound: u32,
}

pub fn existence_checker_stage<RP: RawPile>(
    pile: Arc<Pile<RP>>,
    next_stage: Sender<(ChunkId, Vec<u8>)>,
) -> Sender<(ChunkId, Vec<u8>)> {
    let shared_seen_set: Arc<Mutex<HashSet<ChunkId>>> = Default::default();
    let (tx, rx) = crossbeam_channel::bounded::<(ChunkId, Vec<u8>)>(32);

    // TODO would like something better for the networked case
    for _ in 0..get_number_of_workers("YAMA_EXISTENCE_CHECKERS") {
        let shared_seen_set = shared_seen_set.clone();
        let next_stage = next_stage.clone();
        let rx = rx.clone();
        let pile = pile.clone();
        std::thread::Builder::new()
            .name("yama exist?er".to_string())
            .spawn(move || {
                while let Ok((chunk_id, chunk)) = rx.recv() {
                    // TODO handle errors properly
                    let is_new = { shared_seen_set.lock().unwrap().insert(chunk_id) };
                    if !is_new {
                        continue;
                    }
                    if !pile.chunk_exists(&chunk_id).unwrap() {
                        next_stage.send((chunk_id, chunk)).unwrap();
                    }
                }
            })
            .unwrap();
    }

    tx
}

pub enum ControllerMessage {
    Failure {
        worker_id: Arc<String>,
        error_message: String,
    },
}

pub trait RawPile: Send + Sync + Debug + 'static {
    // TODO expose verification errors?
    fn exists(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<bool>;
    fn read(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<Option<Vec<u8>>>;
    fn write(&self, kind: Keyspace, key: &[u8], value: &[u8]) -> anyhow::Result<()>;
    fn delete(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<()>;
    fn list_keys(
        &self,
        kind: Keyspace,
    ) -> anyhow::Result<Box<dyn Iterator<Item = anyhow::Result<Vec<u8>>>>>;
    /*
    fn list_keyvalue_pairs(
        &self,
        kind: Keyspace,
    ) -> anyhow::Result<Box<dyn Iterator<Item = anyhow::Result<(Vec<u8>, Vec<u8>)>>>>;
    */
    fn flush(&self) -> anyhow::Result<()>;
    // TODO return a progress Receiver
    fn check_lowlevel(&self) -> anyhow::Result<bool>;

    /// Return a few statistics, if possible.
    fn debug_statistics(&self) -> anyhow::Result<Option<DebugStatistics>> {
        Ok(None)
    }

    fn build_storage_pipeline(
        &self,
        settings: StoragePipelineSettings,
        controller_send: Sender<ControllerMessage>,
    ) -> anyhow::Result<Sender<(ChunkId, Vec<u8>)>>;
}

impl RawPile for Box<dyn RawPile> {
    fn exists(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<bool> {
        self.as_ref().exists(kind, key)
    }
    fn read(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<Option<Vec<u8>>> {
        self.as_ref().read(kind, key)
    }
    fn write(&self, kind: Keyspace, key: &[u8], value: &[u8]) -> anyhow::Result<()> {
        self.as_ref().write(kind, key, value)
    }
    fn delete(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<()> {
        self.as_ref().delete(kind, key)
    }
    fn list_keys(
        &self,
        kind: Keyspace,
    ) -> anyhow::Result<Box<dyn Iterator<Item = anyhow::Result<Vec<u8>>>>> {
        self.as_ref().list_keys(kind)
    }
    fn flush(&self) -> anyhow::Result<()> {
        self.as_ref().flush()
    }
    fn check_lowlevel(&self) -> anyhow::Result<bool> {
        self.as_ref().check_lowlevel()
    }
    fn debug_statistics(&self) -> anyhow::Result<Option<DebugStatistics>> {
        self.as_ref().debug_statistics()
    }

    fn build_storage_pipeline(
        &self,
        settings: StoragePipelineSettings,
        controller_send: Sender<ControllerMessage>,
    ) -> anyhow::Result<Sender<(ChunkId, Vec<u8>)>> {
        self.as_ref()
            .build_storage_pipeline(settings, controller_send)
    }
}

impl<RP: RawPile> RawPile for Arc<RP> {
    fn exists(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<bool> {
        self.as_ref().exists(kind, key)
    }
    fn read(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<Option<Vec<u8>>> {
        self.as_ref().read(kind, key)
    }
    fn write(&self, kind: Keyspace, key: &[u8], value: &[u8]) -> anyhow::Result<()> {
        self.as_ref().write(kind, key, value)
    }
    fn delete(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<()> {
        self.as_ref().delete(kind, key)
    }
    fn list_keys(
        &self,
        kind: Keyspace,
    ) -> anyhow::Result<Box<dyn Iterator<Item = anyhow::Result<Vec<u8>>>>> {
        self.as_ref().list_keys(kind)
    }
    fn flush(&self) -> anyhow::Result<()> {
        self.as_ref().flush()
    }
    fn check_lowlevel(&self) -> anyhow::Result<bool> {
        self.as_ref().check_lowlevel()
    }
    fn debug_statistics(&self) -> anyhow::Result<Option<DebugStatistics>> {
        self.as_ref().debug_statistics()
    }

    fn build_storage_pipeline(
        &self,
        settings: StoragePipelineSettings,
        controller_send: Sender<ControllerMessage>,
    ) -> anyhow::Result<Sender<(ChunkId, Vec<u8>)>> {
        self.as_ref()
            .build_storage_pipeline(settings, controller_send)
    }
}

#[derive(Debug)]
pub struct Pile<R: RawPile> {
    pub raw_pile: R,
    pub racy_submission_mutex: Mutex<HashSet<ChunkId>>,
    pub racy_submission_condvar: Condvar,
}

impl<R: RawPile> Pile<R> {
    pub fn new(raw_pile: R) -> Self {
        Pile {
            raw_pile,
            racy_submission_mutex: Mutex::new(Default::default()),
            racy_submission_condvar: Default::default(),
        }
    }

    // TODO(clarity, features): have a special kind of error for verification failures
    //      may be wanted for best-effort restores
    pub fn read_chunk(&self, key: &ChunkId) -> anyhow::Result<Option<Vec<u8>>> {
        self.raw_pile.read(Keyspace::Chunk, key)
        /*
        let result = self.raw_pile.read(Keyspace::Chunk, &key)?;
        if let Some(chunk) = result {
            if verify {
                let hash = self
                    .raw_pile
                    .read(Keyspace::ChunkHash, &key)?
                    .ok_or_else(|| {
                        anyhow!(
                            "Hash not found for chunk {}; can't verify",
                            bytes_to_hexstring(&key)
                        )
                    })?;
                let mut hasher = twox_hash::XxHash64::with_seed(XXH64_SEED);
                hasher.write(&chunk);
                let computed_hash = hasher.finish().to_be_bytes();
                if &computed_hash[..] != &hash {
                    bail!(
                        "Hash mismatch for chunk {}: expected {} computed {}",
                        bytes_to_hexstring(&key),
                        bytes_to_hexstring(&hash),
                        bytes_to_hexstring(&computed_hash),
                    );
                }
            }
            Ok(Some(chunk))
        } else {
            Ok(None)
        }
        */
    }

    pub fn write_chunk(&self, key: &ChunkId, value: &[u8]) -> anyhow::Result<()> {
        self.raw_pile.write(Keyspace::Chunk, key, value)
    }

    pub fn chunk_exists(&self, key: &ChunkId) -> anyhow::Result<bool> {
        self.raw_pile.exists(Keyspace::Chunk, key)
    }

    pub fn read_pointer(&self, key: &str) -> anyhow::Result<Option<PointerData>> {
        Ok(
            if let Some(pointer_data_raw) = self.raw_pile.read(Keyspace::Pointer, key.as_bytes())? {
                Some(serde_bare::from_slice(&pointer_data_raw)?)
            } else {
                None
            },
        )
    }

    pub fn write_pointer(&self, key: &str, pointer: &PointerData) -> anyhow::Result<()> {
        self.raw_pile.write(
            Keyspace::Pointer,
            key.as_bytes(),
            &serde_bare::to_vec(pointer)?,
        )
    }

    pub fn delete_pointer(&self, key: &str) -> anyhow::Result<()> {
        self.raw_pile.delete(Keyspace::Pointer, key.as_bytes())
    }

    pub fn list_pointers(&self) -> anyhow::Result<Vec<String>> {
        let mut result = Vec::new();
        for key in self.raw_pile.list_keys(Keyspace::Pointer)? {
            result.push(String::from_utf8(key?)?);
        }
        Ok(result)
    }

    pub fn submit_chunk(&self, chunk_id: ChunkId, chunk_data: &[u8]) -> anyhow::Result<()> {
        let mut racy_submissions = self.racy_submission_mutex.lock().unwrap();
        if racy_submissions.insert(chunk_id) {
            drop(racy_submissions);
            if !self.chunk_exists(&chunk_id)? {
                self.write_chunk(&chunk_id, chunk_data)?;
            }
            racy_submissions = self.racy_submission_mutex.lock().unwrap();
            racy_submissions.remove(&chunk_id);
            // wake up anyone who might be waiting for this chunk
            self.racy_submission_condvar.notify_all();
        } else {
            loop {
                racy_submissions = self.racy_submission_condvar.wait(racy_submissions).unwrap();
                if !racy_submissions.contains(&chunk_id) {
                    break;
                }
            }
        }
        Ok(())
    }

    /// Flushes buffered writes. Should really run this before exiting, so I can sleep better at
    /// night (rather than relying on the destructor).
    pub fn flush(&self) -> anyhow::Result<()> {
        self.raw_pile.flush()
    }
}
