/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::sync::Arc;
use std::thread;
use std::thread::JoinHandle;

use anyhow::anyhow;
use crossbeam_channel::{Receiver, Sender};
use derivative::Derivative;
use log::error;
use metrics::{register_counter, Unit};
use zstd::block::{Compressor, Decompressor};

use crate::definitions::ChunkId;
use crate::pile::{ControllerMessage, DebugStatistics, Keyspace, RawPile, StoragePipelineSettings};

pub const DECOMPRESS_CAPACITY: usize = 32 * 1024 * 1024;

#[derive(Clone, Debug)]
pub struct CompressionSettings {
    /// Raw dictionary to pass to Zstd for compression and decompression
    pub dictionary: Arc<Vec<u8>>,
    /// The compression level, passed to Zstd.
    pub level: i32,
    /// The number of compressor threads to use.
    pub num_compressors: u32,
    /// The number of decompressor threads to use.
    pub num_decompressors: u32,
}

#[derive(Debug, Derivative)]
#[derivative(Clone(bound = ""))]
// we need to use derivative's Clone impl because Arc<R> causes R to have a bound on Clone
// even though that's not needed. https://github.com/rust-lang/rust/issues/26925
pub struct RawPileCompressor<R: RawPile> {
    underlying: Arc<R>,
    compressor: Option<Sender<(Vec<u8>, Sender<Vec<u8>>)>>,
    decompressor: Option<Sender<(Vec<u8>, Sender<Vec<u8>>)>>,
    settings: Arc<CompressionSettings>,
}

impl<R: RawPile> RawPileCompressor<R> {
    pub fn new(
        underlying: R,
        settings: CompressionSettings,
    ) -> anyhow::Result<(Self, Vec<JoinHandle<()>>)> {
        register_counter!(
            "compressor_in_bytes",
            Unit::Bytes,
            "Number of bytes that have been fed into the compressor"
        );
        register_counter!(
            "compressor_out_bytes",
            Unit::Bytes,
            "Number of bytes that have come out of the compressor"
        );
        register_counter!(
            "compressor_chunks",
            Unit::Count,
            "Number of chunks that have been compressed"
        );
        register_counter!(
            "decompressor_in_bytes",
            Unit::Bytes,
            "Number of bytes that have been fed into the decompressor"
        );
        register_counter!(
            "decompressor_out_bytes",
            Unit::Bytes,
            "Number of bytes that have come out of the decompressor"
        );
        register_counter!(
            "decompressor_chunks",
            Unit::Count,
            "Number of chunks that have been decompressed"
        );

        if settings.num_compressors == 0 && settings.num_decompressors == 0 {
            // optimisation for when we're only building a pipeline: we don't want to
            return Ok((
                RawPileCompressor {
                    underlying: Arc::new(underlying),
                    compressor: None,
                    decompressor: None,

                    settings: Arc::new(settings),
                },
                Vec::with_capacity(0),
            ));
        }

        let (com_s, com_r) = crossbeam_channel::bounded(4);
        let (dec_s, dec_r) = crossbeam_channel::bounded(4);

        let mut handles = Vec::new();

        for worker in 0..settings.num_compressors {
            let settings = settings.clone();
            let com_r = com_r.clone();

            let builder = thread::Builder::new().name(format!("yama compressor {}", worker));

            handles.push(builder.spawn(move || {
                if let Err(e) = Self::compressor_worker(com_r, settings) {
                    error!("compressor worker failed: {:?}", e);
                }
            })?);
        }

        for worker in 0..settings.num_decompressors {
            let settings = settings.clone();
            let dec_r = dec_r.clone();

            let builder = thread::Builder::new().name(format!("yama decompressor {}", worker));

            handles.push(builder.spawn(move || {
                if let Err(e) = Self::decompressor_worker(dec_r, settings) {
                    error!("decompressor worker failed: {:?}", e);
                }
            })?);
        }

        Ok((
            RawPileCompressor {
                underlying: Arc::new(underlying),
                compressor: Some(com_s),
                decompressor: Some(dec_s),
                settings: Arc::new(settings),
            },
            handles,
        ))
    }

    fn compressor_worker(
        queue: Receiver<(Vec<u8>, Sender<Vec<u8>>)>,
        settings: CompressionSettings,
    ) -> anyhow::Result<()> {
        let mut compressor = Compressor::with_dict(settings.dictionary.as_ref().clone());
        while let Ok((job, response_sender)) = queue.recv() {
            let result = compressor.compress(&job, settings.level)?;
            response_sender
                .send(result)
                .or(Err(anyhow!("Couldn't send compression result")))?;
        }
        Ok(())
    }

    fn decompressor_worker(
        queue: Receiver<(Vec<u8>, Sender<Vec<u8>>)>,
        settings: CompressionSettings,
    ) -> anyhow::Result<()> {
        let mut decompressor = Decompressor::with_dict(settings.dictionary.as_ref().clone());
        while let Ok((job, response_sender)) = queue.recv() {
            let result = decompressor.decompress(&job, DECOMPRESS_CAPACITY)?;
            response_sender
                .send(result)
                .or(Err(anyhow!("Couldn't send decompression result")))?;
        }
        Ok(())
    }

    fn decompress(&self, data: &[u8]) -> anyhow::Result<Vec<u8>> {
        let (ret_s, ret_r) = crossbeam_channel::bounded(0);
        self.decompressor
            .as_ref()
            .expect("No decompressors configured")
            .send((data.to_vec(), ret_s))
            .or(Err(anyhow!("couldn't send to decompressor")))?;

        Ok(ret_r.recv().or(Err(anyhow!("couldn't receive result")))?)
    }

    fn compress(&self, compressed_data: &[u8]) -> anyhow::Result<Vec<u8>> {
        let (ret_s, ret_r) = crossbeam_channel::bounded(0);
        self.compressor
            .as_ref()
            .expect("No compressors configured")
            .send((compressed_data.to_vec(), ret_s))
            .or(Err(anyhow!("couldn't send to compressor")))?;

        Ok(ret_r.recv().or(Err(anyhow!("couldn't receive result")))?)
    }

    fn storage_pipeline_worker(
        &self,
        next_stage: Sender<(ChunkId, Vec<u8>)>,
        input: Receiver<(ChunkId, Vec<u8>)>,
        worker_id: String,
    ) -> anyhow::Result<()> {
        // the worker ID has to live forever, so we leak it :/
        let worker_id: &'static str = Box::leak(worker_id.into_boxed_str());
        metrics::register_histogram!(
            "compressor_idle_time",
            metrics::Unit::Seconds,
            "Time spent waiting between chunks",
            "id" => worker_id
        );
        metrics::register_counter!(
            "compressor_bytes_input",
            metrics::Unit::Bytes,
            "Number of bytes input into the compressor.",
            "id" => worker_id
        );
        metrics::register_counter!(
            "compressor_bytes_output",
            metrics::Unit::Bytes,
            "Number of bytes output from the compressor.",
            "id" => worker_id
        );
        metrics::register_counter!(
            "compressor_chunks_processed",
            metrics::Unit::Count,
            "Number of bytes input into the compressor.",
            "id" => worker_id
        );

        let mut compressor = Compressor::with_dict(self.settings.dictionary.as_ref().clone());
        let level = self.settings.level;
        while let Ok((chunk_id, bytes)) = input.recv() {
            let in_bytes = bytes.len();
            let bytes = compressor.compress(&bytes, level)?;
            let out_bytes = bytes.len();
            next_stage.send((chunk_id, bytes))?;
            // Per-worker metrics
            // TODO rename
            metrics::counter!("compressor_bytes_input", in_bytes as u64, "id" => worker_id);
            metrics::counter!("compressor_bytes_output", out_bytes as u64, "id" => worker_id);

            // Global metrics
            metrics::counter!("compressor_in_bytes", in_bytes as u64);
            metrics::counter!("compressor_out_bytes", out_bytes as u64);
            metrics::increment_counter!("compressor_chunks");
        }

        Ok(())
    }
}

impl<R: RawPile> RawPile for RawPileCompressor<R> {
    fn exists(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<bool> {
        self.underlying.exists(kind, key)
    }

    fn read(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<Option<Vec<u8>>> {
        if let Some(data) = self.underlying.read(kind, key)? {
            Ok(Some(self.decompress(&data)?))
        } else {
            Ok(None)
        }
    }

    fn write(&self, kind: Keyspace, key: &[u8], value: &[u8]) -> anyhow::Result<()> {
        let compressed = self.compress(value)?;
        self.underlying.write(kind, key, &compressed)
    }

    fn delete(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<()> {
        self.underlying.delete(kind, key)
    }

    fn list_keys(
        &self,
        kind: Keyspace,
    ) -> anyhow::Result<Box<dyn Iterator<Item = anyhow::Result<Vec<u8>>>>> {
        self.underlying.list_keys(kind)
    }
    fn flush(&self) -> anyhow::Result<()> {
        self.underlying.flush()
    }
    fn check_lowlevel(&self) -> anyhow::Result<bool> {
        self.underlying.check_lowlevel()
    }

    fn debug_statistics(&self) -> anyhow::Result<Option<DebugStatistics>> {
        self.underlying.debug_statistics()
    }

    fn build_storage_pipeline(
        &self,
        settings: StoragePipelineSettings,
        controller_send: Sender<ControllerMessage>,
    ) -> anyhow::Result<Sender<(ChunkId, Vec<u8>)>> {
        // this one should have a few threads behind it! yarr!
        let subsequent_pipeline = self
            .underlying
            .build_storage_pipeline(settings.clone(), controller_send.clone())?;

        let (input_to_this_stage, receiver) =
            crossbeam_channel::bounded(settings.compressor_input_bound as usize);

        for compressor_number in 0..settings.num_compressors {
            let subsequent_pipeline = subsequent_pipeline.clone();
            let receiver = receiver.clone();
            let controller_send = controller_send.clone();
            let this = (*self).clone();
            thread::Builder::new()
                .name(format!("yama Pcomp{}", compressor_number))
                .spawn(move || {
                    let worker_id = Arc::new(format!("compressor-{}", compressor_number));
                    if let Err(err) = this.storage_pipeline_worker(
                        subsequent_pipeline,
                        receiver,
                        worker_id.to_string(),
                    ) {
                        controller_send
                            .send(ControllerMessage::Failure {
                                worker_id,
                                error_message: format!("err {:?}", err),
                            })
                            .expect("This is BAD: failed to send failure message to controller.");
                    }
                })
                .unwrap();
        }

        Ok(input_to_this_stage)
    }
}
