/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use anyhow::anyhow;
use log::warn;
use sodiumoxide::crypto::secretbox;
use sodiumoxide::crypto::secretbox::{Key, Nonce, NONCEBYTES};

use crate::definitions::ChunkId;
use crate::pile::{ControllerMessage, Keyspace, RawPile, StoragePipelineSettings};
use crossbeam_channel::Sender;

/// A RawPile that provides encryption of chunk contents.
/// Please note that keys are not currently encrypted, so this scheme is not CPA-secure.
/// It seems easily possible to test the pile for inclusion of a known file (by first chunking it and
/// looking for matching chunk IDs).
/// Use of compression a custom Zstd dictionary may make that harder but in general it seems dubious
/// to rely on that.
/// This feature will be revisited soon...
/// Notably, keys should be passed through a secure permutation first.
#[derive(Debug)]
pub struct RawPileEncryptor<R: RawPile> {
    underlying: R,
    secret_key: Key,
}

impl<R: RawPile> RawPileEncryptor<R> {
    pub fn new(underlying: R, key: Key) -> Self {
        warn!(
            "WARNING! Encrypted RawPiles are not CPA secure. Do not rely on them for security yet!"
        );
        RawPileEncryptor {
            underlying,
            secret_key: key,
        }
    }

    fn decrypt(&self, kind: Keyspace, key: &[u8], data: &[u8]) -> anyhow::Result<Vec<u8>> {
        Ok(if kind == Keyspace::Chunk {
            let mut nonce = [0u8; NONCEBYTES];
            nonce[0..key.len()].copy_from_slice(key);
            secretbox::open(data, &Nonce(nonce), &self.secret_key)
                .or(Err(anyhow!("Failed to decrypt")))?
        } else {
            let mut nonce = [0u8; NONCEBYTES];
            nonce.copy_from_slice(&data[0..NONCEBYTES]);
            secretbox::open(&data[NONCEBYTES..], &Nonce(nonce), &self.secret_key)
                .or(Err(anyhow!("Failed to decrypt")))?
        })
    }

    fn encrypt(&self, kind: Keyspace, key: &[u8], data: &[u8]) -> Vec<u8> {
        if kind == Keyspace::Chunk {
            let mut nonce = [0u8; NONCEBYTES];
            nonce[0..key.len()].copy_from_slice(key);
            secretbox::seal(data, &Nonce(nonce), &self.secret_key)
        } else {
            let nonce = secretbox::gen_nonce();
            let mut out = Vec::new();
            out.extend_from_slice(&nonce.0);
            out.extend_from_slice(&secretbox::seal(data, &nonce, &self.secret_key));
            out
        }
    }
}

impl<R: RawPile> RawPile for RawPileEncryptor<R> {
    fn exists(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<bool> {
        self.underlying.exists(kind, key)
    }

    fn read(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<Option<Vec<u8>>> {
        if let Some(data) = self.underlying.read(kind, key)? {
            Ok(Some(self.decrypt(kind, key, &data)?))
        } else {
            Ok(None)
        }
    }
    fn write(&self, kind: Keyspace, key: &[u8], value: &[u8]) -> anyhow::Result<()> {
        let encrypted = self.encrypt(kind, key, value);
        self.underlying.write(kind, key, &encrypted)
    }

    fn delete(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<()> {
        self.underlying.delete(kind, key)
    }

    fn list_keys(
        &self,
        kind: Keyspace,
    ) -> anyhow::Result<Box<dyn Iterator<Item = anyhow::Result<Vec<u8>>>>> {
        self.underlying.list_keys(kind)
    }
    fn flush(&self) -> anyhow::Result<()> {
        self.underlying.flush()
    }
    fn check_lowlevel(&self) -> anyhow::Result<bool> {
        self.underlying.check_lowlevel()
    }

    fn build_storage_pipeline(
        &self,
        settings: StoragePipelineSettings,
        controller_send: Sender<ControllerMessage>,
    ) -> anyhow::Result<Sender<(ChunkId, Vec<u8>)>> {
        todo!()
    }
}
