/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::convert::TryInto;
use std::io::{Read, Write};

use byteorder::{BigEndian, ReadBytesExt, WriteBytesExt};
use serde::de::DeserializeOwned;
use serde::{Deserialize, Serialize};

use crate::pile::Keyspace;

pub mod requester;
pub mod responder;

#[derive(Serialize, Deserialize, Clone)]
pub struct Request {
    id: u16,
    body: RequestBody,
}

#[derive(Serialize, Deserialize, Clone, Eq, PartialEq)]
pub enum RequestBody {
    Read {
        kind: Keyspace,
        key: Vec<u8>,
    },
    CheckExists {
        kind: Keyspace,
        key: Vec<u8>,
    },
    Write {
        kind: Keyspace,
        key: Vec<u8>,
        value: Vec<u8>,
    },
    Delete {
        kind: Keyspace,
        key: Vec<u8>,
    },
    List {
        kind: Keyspace,
    },
    NextBatch {
        token: u16,
    },
    Flush,
    LowLevelCheck,
    Shutdown,
    Progress {
        current: u64,
        max: u64,
    },
}

#[derive(Serialize, Deserialize, Clone)]
pub struct Response {
    response_to: u16,
    body: ResponseBody,
}

#[derive(Serialize, Deserialize, Clone)]
pub enum ResponseBody {
    Success,
    Failed(String),
    NotExists,
    Data(Vec<u8>),
    BatchData {
        batch: Vec<Vec<u8>>,
        next_token: u16,
    },
}

pub fn read_message<R: Read, D: DeserializeOwned>(read: &mut R) -> anyhow::Result<D> {
    let len = read.read_u32::<BigEndian>()?;
    let mut data_vec = vec![0u8; len as usize];
    read.read_exact(&mut data_vec)?;
    Ok(serde_bare::from_slice(&data_vec)?)
}

pub fn write_message<W: Write, S: Serialize>(write: &mut W, message: &S) -> anyhow::Result<()> {
    let data_vec = serde_bare::to_vec(&message)?;
    write.write_u32::<BigEndian>(data_vec.len().try_into()?)?;
    write.write_all(&data_vec)?;
    Ok(())
}
