use std::collections::HashMap;
use std::io::{stdin, stdout, Read, Write};
use std::sync::{Arc, Mutex};
use std::thread;
use std::thread::JoinHandle;

use anyhow::anyhow;
use crossbeam_channel::{Receiver, Sender};
use log::{error, info};

use crate::definitions::ChunkId;
use crate::pile::{ControllerMessage, Keyspace, RawPile, StoragePipelineSettings};
use crate::remote::{read_message, write_message, Request, RequestBody, Response, ResponseBody};
use std::sync::atomic::{AtomicBool, AtomicU16, Ordering};

/// A kind of RawPile which can make requests to a RawPile over a pipe (e.g. TCP socket or an
/// SSH connection).
/// The requests are handled by a `Responder` on the other end of the pipe.
#[derive(Debug)]
pub struct Requester {
    commands: Sender<(RequestBody, Option<Sender<ResponseBody>>)>,
}

impl Requester {
    pub fn new<R: Read + Send + 'static, W: Write + Send + 'static>(
        read: R,
        write: W,
    ) -> (Self, Vec<JoinHandle<()>>) {
        let in_flight: Arc<Mutex<HashMap<u16, Sender<ResponseBody>>>> =
            Arc::new(Mutex::new(HashMap::new()));
        let (command_sender, command_receiver) = crossbeam_channel::bounded(16);
        let mut handles = Vec::new();

        let shutdown_signal: Arc<(AtomicU16, AtomicBool)> = Default::default();

        {
            // Spawn a reader
            let in_flight = in_flight.clone();
            let shutdown_signal = shutdown_signal.clone();
            handles.push(thread::spawn(move || {
                if let Err(e) = Self::reader(read, in_flight, shutdown_signal) {
                    error!("reader failed: {:?}", e);
                }
            }));
        }

        {
            // Spawn a writer
            let in_flight = in_flight.clone();
            let command_receiver = command_receiver.clone();
            handles.push(thread::spawn(move || {
                if let Err(e) = Self::writer(write, in_flight, command_receiver, shutdown_signal) {
                    error!("writer failed: {:?}", e);
                }
            }));
        }

        (
            Requester {
                commands: command_sender,
            },
            handles,
        )
    }

    pub fn new_from_stdio() -> (Self, Vec<JoinHandle<()>>) {
        let in_flight: Arc<Mutex<HashMap<u16, Sender<ResponseBody>>>> =
            Arc::new(Mutex::new(HashMap::new()));
        let (command_sender, command_receiver) = crossbeam_channel::bounded(16);
        let mut handles = Vec::new();

        let shutdown_signal: Arc<(AtomicU16, AtomicBool)> = Default::default();

        {
            // Spawn a reader
            let in_flight = in_flight.clone();
            let shutdown_signal = shutdown_signal.clone();
            handles.push(thread::spawn(move || {
                let stdin = stdin();
                let read = stdin.lock();
                if let Err(e) = Self::reader(read, in_flight, shutdown_signal) {
                    error!("reader failed: {:?}", e);
                }
            }));
        }

        {
            // Spawn a writer
            let in_flight = in_flight.clone();
            let command_receiver = command_receiver.clone();
            handles.push(thread::spawn(move || {
                let stdout = stdout();
                let write = stdout.lock();
                if let Err(e) = Self::writer(write, in_flight, command_receiver, shutdown_signal) {
                    error!("writer failed: {:?}", e);
                }
            }));
        }

        (
            Requester {
                commands: command_sender,
            },
            handles,
        )
    }

    pub fn clone_command_sender(&self) -> Sender<(RequestBody, Option<Sender<ResponseBody>>)> {
        self.commands.clone()
    }

    /// Thread that reads messages and sends them along.
    fn reader<R: Read>(
        mut read: R,
        in_flight: Arc<Mutex<HashMap<u16, Sender<ResponseBody>>>>,
        shutdown_request_channel: Arc<(AtomicU16, AtomicBool)>,
    ) -> anyhow::Result<()> {
        loop {
            let response: Response = read_message(&mut read)?;

            if shutdown_request_channel.1.load(Ordering::Relaxed)
                && response.response_to == shutdown_request_channel.0.load(Ordering::Relaxed)
            {
                return Ok(());
            }

            let map = in_flight.lock().or(Err(anyhow!("Mutex poisoned")))?;
            map.get(&response.response_to)
                .ok_or(anyhow!("Didn't find response channel..."))?
                .send(response.body)
                .or(Err(anyhow!("Failed to send response to channel")))?;
        }
    }

    /// Thread that writes messages.
    fn writer<W: Write>(
        mut write: W,
        in_flight: Arc<Mutex<HashMap<u16, Sender<ResponseBody>>>>,
        command_receiver: Receiver<(RequestBody, Option<Sender<ResponseBody>>)>,
        shutdown_request_channel: Arc<(AtomicU16, AtomicBool)>,
    ) -> anyhow::Result<()> {
        while let Ok((req_body, response_channel)) = command_receiver.recv() {
            let request_id = if let Some(response_channel) = response_channel {
                let mut map = in_flight.lock().or(Err(anyhow!("Mutex poisoned")))?;
                let request_id = (1u16..u16::MAX)
                    .into_iter()
                    .find(|id| !map.contains_key(&id))
                    .expect("No ID found");
                map.insert(request_id, response_channel);
                request_id
            } else {
                0
            };

            let shutting_down = &req_body == &RequestBody::Shutdown;

            if shutting_down {
                shutdown_request_channel
                    .0
                    .store(request_id, Ordering::SeqCst);
                shutdown_request_channel.1.store(true, Ordering::SeqCst);
            }

            write_message(
                &mut write,
                &Request {
                    id: request_id,
                    body: req_body,
                },
            )?;
            write.flush()?;

            if shutting_down {
                return Ok(());
            }
        }

        info!("Exited send loop without shutdown message, issuing Shutdown.");
        // shutdown ourselves
        let request_id = {
            let map = in_flight.lock().or(Err(anyhow!("Mutex poisoned")))?;
            let request_id = (0u16..u16::MAX)
                .into_iter()
                .find(|id| !map.contains_key(&id))
                .expect("No ID found");
            request_id
        };

        shutdown_request_channel
            .0
            .store(request_id, Ordering::SeqCst);
        shutdown_request_channel.1.store(true, Ordering::SeqCst);

        write_message(
            &mut write,
            &Request {
                id: request_id,
                body: RequestBody::Shutdown,
            },
        )?;
        write.flush()?;

        Ok(())
    }

    /// Helper to make a request and wait for the result.
    fn request(&self, req: RequestBody) -> anyhow::Result<ResponseBody> {
        let (sender, receiver) = crossbeam_channel::bounded(0);
        self.commands
            .send((req, Some(sender)))
            .or(Err(anyhow!("Failed to queue request")))?;

        Ok(receiver
            .recv()
            .or(Err(anyhow!("Failed to receive response")))?)
    }
}

impl RawPile for Requester {
    fn exists(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<bool> {
        match self.request(RequestBody::CheckExists {
            kind,
            key: key.to_vec(),
        })? {
            ResponseBody::Success => Ok(true),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Ok(false),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for exists.")),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for exists.")),
        }
    }
    fn read(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<Option<Vec<u8>>> {
        match self.request(RequestBody::Read {
            kind,
            key: key.to_vec(),
        })? {
            ResponseBody::Success => Err(anyhow!("Received Success for read.")),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Ok(None),
            ResponseBody::Data(data) => Ok(Some(data)),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for read.")),
        }
    }
    fn write(&self, kind: Keyspace, key: &[u8], value: &[u8]) -> anyhow::Result<()> {
        match self.request(RequestBody::Write {
            kind,
            key: key.to_vec(),
            value: value.to_vec(),
        })? {
            ResponseBody::Success => Ok(()),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Err(anyhow!("Received NotExists for write.")),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for write.")),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for write.")),
        }
    }
    fn delete(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<()> {
        match self.request(RequestBody::Delete {
            kind,
            key: key.to_vec(),
        })? {
            ResponseBody::Success => Ok(()),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Err(anyhow!("Received NotExists for delete.")),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for delete.")),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for delete.")),
        }
    }
    fn list_keys(
        &self,
        kind: Keyspace,
    ) -> anyhow::Result<Box<dyn Iterator<Item = anyhow::Result<Vec<u8>>>>> {
        match self.request(RequestBody::List { kind })? {
            ResponseBody::Success => Ok(Box::new(ListKeyIterator {
                command_sender: self.commands.clone(),
                batch_token: None,
                buffer: Vec::with_capacity(0),
            })),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Err(anyhow!("Received NotExists for list_keys.")),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for list_keys.")),
            ResponseBody::BatchData { batch, next_token } => Ok(Box::new(ListKeyIterator {
                command_sender: self.commands.clone(),
                batch_token: Some(next_token),
                buffer: batch,
            })),
        }
    }
    fn flush(&self) -> anyhow::Result<()> {
        match self.request(RequestBody::Flush)? {
            ResponseBody::Success => Ok(()),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Err(anyhow!("Received NotExists for Flush.")),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for Flush.")),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for Flush.")),
        }
    }
    fn check_lowlevel(&self) -> anyhow::Result<bool> {
        match self.request(RequestBody::LowLevelCheck)? {
            ResponseBody::Success => Ok(true),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Err(anyhow!("Received NotExists for LowLevelCheck.")),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for LowLevelCheck.")),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for LowLevelCheck.")),
        }
    }

    fn build_storage_pipeline(
        &self,
        settings: StoragePipelineSettings,
        controller_send: Sender<ControllerMessage>,
    ) -> anyhow::Result<Sender<(ChunkId, Vec<u8>)>> {
        // this one is a little bit more complex.
        // We want to be able to send off multiple write requests at once, but not too many, so we
        // need to be able to apply backpressure.
        let (input, receiver) = crossbeam_channel::bounded::<(ChunkId, Vec<u8>)>(128);
        let command_sender = self.commands.clone();

        std::thread::spawn(move || {
            let (response_tx, response_rx) = crossbeam_channel::bounded::<ResponseBody>(32);
            let mut in_flight_writes = 0;
            const MAX_IN_FLIGHT_WRITES: u32 = 32;
            let mut pipeline_still_going = true;

            while pipeline_still_going || in_flight_writes > 0 {
                // TODO this won't handle channel closure properly.
                if in_flight_writes < MAX_IN_FLIGHT_WRITES && pipeline_still_going {
                    crossbeam_channel::select! {
                        recv(response_rx) -> resp => {
                            in_flight_writes -= 1;
                            match resp.unwrap() {
                                ResponseBody::Success => {
                                    // nop
                                }
                                ResponseBody::Failed(string) => {
                                    panic!("Requester pipeline fail {}", string);
                                }
                                ResponseBody::BatchData { .. } => {
                                    panic!("wtf BatchData");
                                }
                                ResponseBody::NotExists => {
                                    panic!("wtf NotExists");
                                }
                                ResponseBody::Data(_) => {
                                    panic!("wtf Data");
                                }
                            }
                        }
                        recv(receiver) -> resp => {
                            if let Ok((chunk_id, write)) = resp {
                                in_flight_writes += 1;
                                command_sender.send((RequestBody::Write {
                                    kind: Keyspace::Chunk,
                                    key: chunk_id.to_vec(),
                                    value: write
                                }, Some(response_tx.clone()))).unwrap();
                            } else {
                                // the input has stopped
                                pipeline_still_going = false;
                            }
                        }
                    }
                } else {
                    // Either the pipeline is stopping or we are too busy to accept new chunks,
                    // so only process responses.
                    let resp = response_rx.recv().unwrap();
                    match resp {
                        ResponseBody::Success => {
                            // nop
                        }
                        ResponseBody::Failed(string) => {
                            panic!("Requester pipeline fail {}", string);
                        }
                        ResponseBody::BatchData { .. } => {
                            panic!("wtf BatchData");
                        }
                        ResponseBody::NotExists => {
                            panic!("wtf NotExists");
                        }
                        ResponseBody::Data(_) => {
                            panic!("wtf Data");
                        }
                    }
                }
            }
        });

        Ok(input)
    }
}

pub struct ListKeyIterator {
    pub(crate) command_sender: Sender<(RequestBody, Option<Sender<ResponseBody>>)>,
    pub(crate) batch_token: Option<u16>,
    /// in reverse order
    pub(crate) buffer: Vec<Vec<u8>>,
}

impl Iterator for ListKeyIterator {
    type Item = anyhow::Result<Vec<u8>>;

    fn next(&mut self) -> Option<Self::Item> {
        if let Some(next) = self.buffer.pop() {
            Some(Ok(next))
        } else if let Some(batch_token) = self.batch_token {
            let (send, recv) = crossbeam_channel::bounded(0);
            self.command_sender
                .send((RequestBody::NextBatch { token: batch_token }, Some(send)))
                .expect("Unable to send");
            let resp = recv.recv().expect("Unable to recv");
            match resp {
                ResponseBody::Success => {
                    self.batch_token = None;
                    None
                }
                ResponseBody::Failed(err_msg) => Some(Err(anyhow!("Remote failure: {}", err_msg))),
                ResponseBody::NotExists => Some(Err(anyhow!("Received NotExists for NextBatch."))),
                ResponseBody::Data(_) => Some(Err(anyhow!("Received Data for NextBatch."))),
                ResponseBody::BatchData { batch, next_token } => {
                    self.batch_token = Some(next_token);
                    self.buffer = batch;
                    self.buffer.reverse();

                    if let Some(next) = self.buffer.pop() {
                        Some(Ok(next))
                    } else {
                        None
                    }
                }
            }
        } else {
            None
        }
    }
}
