use std::collections::HashMap;
use std::io::{Read, Write};
use std::sync::{Arc, Mutex};
use std::thread;
use std::thread::JoinHandle;

use anyhow::anyhow;
use crossbeam_channel::{Receiver, Sender};
use itertools::Itertools;
use log::{error, info, warn};

use crate::definitions::ChunkId;
use crate::pile::{Keyspace, RawPile};
use crate::progress::ProgressTracker;
use crate::remote::{read_message, write_message, Request, RequestBody, Response, ResponseBody};

#[derive(Clone)]
pub struct ResponderWritingPipeline {
    pub pipeline_submission: Sender<(ChunkId, Vec<u8>)>,
}

#[derive(Clone)]
/// A wrapper for a RawPile which allows a `Requester` to access it over a pipe (e.g. TCP socket or
/// an SSH connection).
pub struct Responder {
    continuation_tokens: Arc<Mutex<HashMap<u16, Sender<u16>>>>,
    writing_pipeline: Option<ResponderWritingPipeline>,
}

impl Responder {
    /// Start a 'responder' (command processor).
    pub fn start<
        RP: RawPile + 'static,
        R: Read + Send + 'static,
        W: Write + Send + 'static,
        PT: ProgressTracker + Send + 'static,
    >(
        read: R,
        write: W,
        num_workers: u16,
        pile: Arc<RP>,
        writing_pipeline: Option<ResponderWritingPipeline>,
        mut progress_bar: PT,
    ) -> (JoinHandle<R>, JoinHandle<W>, Vec<JoinHandle<()>>) {
        let mut handles = Vec::new();
        let (work_queue_send, work_queue_recv) = crossbeam_channel::bounded::<Request>(16);
        let (resp_send, resp_recv) = crossbeam_channel::bounded::<Response>(4);
        let responder = Responder {
            continuation_tokens: Arc::new(Mutex::new(Default::default())),
            writing_pipeline,
        };

        let r_handle = {
            // spawn the reader
            let work_queue_send = work_queue_send.clone();
            let responder = responder.clone();
            thread::spawn(move || {
                let mut read = read;
                if let Err(e) = responder.reader(&mut read, work_queue_send, &mut progress_bar) {
                    error!("reader failed: {:?}", e);
                }
                read
            })
        };

        let w_handle = {
            // spawn the writer
            let resp_recv = resp_recv.clone();
            let responder = responder.clone();
            thread::spawn(move || {
                let mut write = write;
                if let Err(e) = responder.writer(&mut write, resp_recv) {
                    error!("writer failed: {:?}", e);
                }
                write
            })
        };

        for worker_num in 0..num_workers {
            // spawn a worker
            let responder = responder.clone();
            let work_queue_recv = work_queue_recv.clone();
            let resp_send = resp_send.clone();
            let pile = pile.clone();
            handles.push(thread::spawn(move || {
                if let Err(e) = responder.worker(pile.as_ref(), work_queue_recv, resp_send) {
                    error!("worker {} failed: {:?}", worker_num, e);
                }
            }));
        }

        (r_handle, w_handle, handles)
    }

    /// Thread that reads messages and sends them along.
    fn reader<R: Read + Send + 'static, PT: ProgressTracker>(
        &self,
        read: &mut R,
        worker_queue_send: Sender<Request>,
        progress_tracker: &mut PT,
    ) -> anyhow::Result<()> {
        loop {
            let request: Request = read_message(read)?;

            match request.body {
                RequestBody::Shutdown => {
                    worker_queue_send.send(request)?;
                    info!("Shutting down responder");
                    return Ok(());
                }
                RequestBody::NextBatch { token } => {
                    let tokens = self
                        .continuation_tokens
                        .lock()
                        .or(Err(anyhow!("Mutex poisoned")))?;
                    tokens
                        .get(&token)
                        .ok_or(anyhow!("Could not find that token."))?
                        .send(request.id)
                        .or(Err(anyhow!(
                            "Failed to send continuation token to continuer"
                        )))?;
                }
                RequestBody::Progress { current, max } => {
                    progress_tracker.set_max_size(max);
                    progress_tracker.set_current(current);
                }
                _ => {
                    worker_queue_send.send(request)?;
                }
            }
        }
    }

    /// Thread that writes messages.
    fn writer<W: Write + Send>(
        &self,
        mut write: W,
        responses: Receiver<Response>,
    ) -> anyhow::Result<()> {
        while let Ok(response) = responses.recv() {
            write_message(&mut write, &response)?;
            write.flush()?;
        }
        Ok(())
    }

    /// Thread which performs the actual work using the pile.
    fn worker<RP: RawPile>(
        &self,
        pile: &RP,
        requests: Receiver<Request>,
        responses: Sender<Response>,
    ) -> anyhow::Result<()> {
        while let Ok(request) = requests.recv() {
            let response = match request.body {
                RequestBody::Read { kind, key } => match pile.read(kind, &key) {
                    Ok(Some(data)) => Response {
                        response_to: request.id,
                        body: ResponseBody::Data(data),
                    },
                    Ok(None) => Response {
                        response_to: request.id,
                        body: ResponseBody::NotExists,
                    },
                    Err(err) => {
                        warn!("Error whilst doing a raw read: {:?}", err);
                        let err = format!("{:?}", err);
                        Response {
                            response_to: request.id,
                            body: ResponseBody::Failed(err),
                        }
                    }
                },
                RequestBody::CheckExists { kind, key } => match pile.exists(kind, &key) {
                    Ok(true) => Response {
                        response_to: request.id,
                        body: ResponseBody::Success,
                    },
                    Ok(false) => Response {
                        response_to: request.id,
                        body: ResponseBody::NotExists,
                    },
                    Err(err) => {
                        warn!("Error whilst doing a raw exists: {:?}", err);
                        let err = format!("{:?}", err);
                        Response {
                            response_to: request.id,
                            body: ResponseBody::Failed(err),
                        }
                    }
                },
                RequestBody::Write { kind, key, value } => {
                    if let Some(writing_pipeline) = self
                        .writing_pipeline
                        .as_ref()
                        .filter(|_| kind == Keyspace::Chunk)
                    {
                        let mut chunk_id = ChunkId::default();
                        chunk_id.copy_from_slice(&key[..]);
                        writing_pipeline
                            .pipeline_submission
                            .send((chunk_id, value))?;
                        // We lie and say it was successful once we submit.
                        // We'll complain on our side if anything goes wrong, anyway.
                        Response {
                            response_to: request.id,
                            body: ResponseBody::Success,
                        }
                    } else {
                        match pile.write(kind, &key, &value) {
                            Ok(_) => Response {
                                response_to: request.id,
                                body: ResponseBody::Success,
                            },
                            Err(err) => {
                                warn!("Error whilst doing a raw write: {:?}", err);
                                let err = format!("{:?}", err);
                                Response {
                                    response_to: request.id,
                                    body: ResponseBody::Failed(err),
                                }
                            }
                        }
                    }
                }
                RequestBody::Delete { kind, key } => match pile.delete(kind, &key) {
                    Ok(_) => Response {
                        response_to: request.id,
                        body: ResponseBody::Success,
                    },
                    Err(err) => {
                        warn!("Error whilst doing a raw delete: {:?}", err);
                        let err = format!("{:?}", err);
                        Response {
                            response_to: request.id,
                            body: ResponseBody::Failed(err),
                        }
                    }
                },
                RequestBody::List { kind } => match pile.list_keys(kind) {
                    Ok(iterator) => {
                        let mut resp_id = request.id;

                        let (cont_sender, cont_receiver) = crossbeam_channel::bounded(1);

                        let batch_token = {
                            let mut map = self
                                .continuation_tokens
                                .lock()
                                .or(Err(anyhow!("Mutex poisoned")))?;
                            let batch_token = (0u16..u16::max_value())
                                .into_iter()
                                .find(|id| !map.contains_key(&id))
                                .expect("No ID found");
                            map.insert(batch_token, cont_sender);
                            batch_token
                        };

                        for chunk in &iterator.chunks(32) {
                            let mut results = Vec::with_capacity(32);
                            for result in chunk {
                                results.push(result?);
                            }

                            responses
                                .send(Response {
                                    response_to: resp_id,
                                    body: ResponseBody::BatchData {
                                        batch: results,
                                        next_token: batch_token,
                                    },
                                })
                                .or(Err(anyhow!("Failed to queue response")))?;

                            resp_id = cont_receiver
                                .recv()
                                .or(Err(anyhow!("Failed to receive continuation response ID")))?;
                        }

                        let mut map = self
                            .continuation_tokens
                            .lock()
                            .or(Err(anyhow!("Mutex poisoned")))?;
                        map.remove(&batch_token);

                        Response {
                            response_to: resp_id,
                            body: ResponseBody::Success,
                        }
                    }
                    Err(err) => {
                        warn!("Error whilst doing a raw list_keys: {:?}", err);
                        let err = format!("{:?}", err);
                        Response {
                            response_to: request.id,
                            body: ResponseBody::Failed(err),
                        }
                    }
                },
                RequestBody::NextBatch { .. } => {
                    unreachable!("because this is handled by the reader")
                }
                RequestBody::Flush => match pile.flush() {
                    Ok(_) => Response {
                        response_to: request.id,
                        body: ResponseBody::Success,
                    },
                    Err(err) => {
                        warn!("Error whilst doing a raw flush: {:?}", err);
                        let err = format!("{:?}", err);
                        Response {
                            response_to: request.id,
                            body: ResponseBody::Failed(err),
                        }
                    }
                },
                RequestBody::LowLevelCheck => match pile.check_lowlevel() {
                    Ok(_) => Response {
                        response_to: request.id,
                        body: ResponseBody::Success,
                    },
                    Err(err) => {
                        warn!("Error whilst doing a raw check_lowlevel: {:?}", err);
                        let err = format!("{:?}", err);
                        Response {
                            response_to: request.id,
                            body: ResponseBody::Failed(err),
                        }
                    }
                },
                RequestBody::Shutdown => Response {
                    response_to: request.id,
                    body: ResponseBody::Success,
                },
                RequestBody::Progress { .. } => {
                    unreachable!("handled by readea")
                }
            };

            responses
                .send(response)
                .or(Err(anyhow!("Failed to queue response")))?;
        }
        Ok(())
    }
}
