/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::collections::btree_map::Entry;
use std::collections::BTreeMap;
use std::fs::{read_link, symlink_metadata, DirEntry, Metadata};
use std::io::ErrorKind;
use std::os::unix::fs::MetadataExt;
use std::path::Path;

use anyhow::anyhow;
use indicatif::{ProgressBar, ProgressDrawTarget, ProgressStyle};
use log::warn;

use crate::definitions::{FilesystemOwnership, FilesystemPermissions, RecursiveChunkRef, TreeNode};

/// Given a file's metadata, returns the mtime in milliseconds.
pub fn mtime_msec(metadata: &Metadata) -> u64 {
    (metadata.mtime() * 1000 + metadata.mtime_nsec() / 1_000_000) as u64
}

/// Scan the filesystem to produce a Tree, using a default progress bar.
pub fn scan(path: &Path) -> anyhow::Result<Option<TreeNode>> {
    let pbar = ProgressBar::with_draw_target(0, ProgressDrawTarget::stdout_with_hz(2));
    pbar.set_style(ProgressStyle::default_spinner().template("{spinner} {pos:7} {msg}"));
    pbar.set_message("dir scan");

    let result = scan_with_progress_bar(path, &pbar);
    pbar.finish_at_current_pos();
    result
}

/// Scan the filesystem to produce a Tree, using the specified progress bar.
pub fn scan_with_progress_bar(
    path: &Path,
    progress_bar: &ProgressBar,
) -> anyhow::Result<Option<TreeNode>> {
    let metadata_res = symlink_metadata(path);
    progress_bar.inc(1);
    if let Err(e) = &metadata_res {
        match e.kind() {
            ErrorKind::NotFound => {
                warn!("vanished: {:?}", path);
                return Ok(None);
            }
            ErrorKind::PermissionDenied => {
                warn!("permission denied: {:?}", path);
                return Ok(None);
            }
            _ => { /* nop */ }
        }
    }
    let metadata = metadata_res?;
    let filetype = metadata.file_type();

    /*let name = path
    .file_name()
    .ok_or(anyhow!("No filename, wat"))?
    .to_str()
    .ok_or(anyhow!("Filename can't be to_str()d"))?
    .to_owned();*/

    let ownership = FilesystemOwnership {
        uid: metadata.uid() as u16,
        gid: metadata.gid() as u16,
    };

    let permissions = FilesystemPermissions {
        mode: metadata.mode(),
    };

    if filetype.is_file() {
        // Leave an unpopulated file node. It's not my responsibility to chunk it right now.
        Ok(Some(TreeNode::NormalFile {
            mtime: mtime_msec(&metadata),
            ownership,
            permissions,
            content: RecursiveChunkRef {
                chunk_id: [0; 32],
                depth: 0,
            },
        }))
    } else if filetype.is_dir() {
        let mut children = BTreeMap::new();
        progress_bar.set_message(&format!("{:?}", path));
        let dir_read = path.read_dir();

        if let Err(e) = &dir_read {
            match e.kind() {
                ErrorKind::NotFound => {
                    warn!("vanished/: {:?}", path);
                    return Ok(None);
                }
                ErrorKind::PermissionDenied => {
                    warn!("permission denied/: {:?}", path);
                    return Ok(None);
                }
                _ => { /* nop */ }
            }
        }

        for entry in dir_read? {
            let entry: DirEntry = entry?;
            let scanned = scan_with_progress_bar(&entry.path(), progress_bar)?;
            if let Some(scanned) = scanned {
                children.insert(
                    entry
                        .file_name()
                        .into_string()
                        .expect("OsString not String"),
                    scanned,
                );
            }
        }

        Ok(Some(TreeNode::Directory {
            ownership,
            permissions,
            children,
        }))
    } else if filetype.is_symlink() {
        let target = read_link(path)?
            .to_str()
            .ok_or(anyhow!("target path cannot be to_str()d"))?
            .to_owned();

        Ok(Some(TreeNode::SymbolicLink { ownership, target }))
    } else {
        Ok(None)
    }
}

/// Differentiates a node in place.
/// This makes `old` the parent of `new` (though it is up to the caller to properly update the
/// `PointerData` to reflect this!).
/// Loosely speaking, `new` is modified to contain the differences that, when applied to `old`, will
/// result in the original value of `new`.
/// See `integrate_node_in_place` for the inverse of this operation.
pub fn differentiate_node_in_place(new: &mut TreeNode, old: &TreeNode) -> anyhow::Result<()> {
    if let TreeNode::Directory { children, .. } = new {
        if let TreeNode::Directory {
            children: old_children,
            ..
        } = old
        {
            for (name, old_node) in old_children.iter() {
                match children.entry(name.clone()) {
                    Entry::Vacant(vacant_entry) => {
                        vacant_entry.insert(TreeNode::Deleted);
                    }
                    Entry::Occupied(occupied_entry) => {
                        if !occupied_entry.get().metadata_invalidates(old_node)? {
                            // The entry is the same, so we don't need to store it!
                            occupied_entry.remove_entry();
                        } else {
                            differentiate_node_in_place(occupied_entry.into_mut(), old_node)?;
                        }
                    }
                }
            }
        }
    }
    Ok(())
}

/// Integrates a node in place.
/// This makes `new` no longer have a parent (remember, the caller is responsible for updating
/// `PointerData` appropriately if needed to reflect this).
///
/// Loosely speaking, `new` is treated as a set of differences that are applied to `old`, though the
/// result is in-place.
///
/// Preconditions:
/// - `old` must be an integrated pointer.
/// - `old` is the parent of `new`
pub fn integrate_node_in_place(new: &mut TreeNode, old: &TreeNode) -> anyhow::Result<()> {
    if let TreeNode::Directory { children, .. } = new {
        if let TreeNode::Directory {
            children: old_children,
            ..
        } = old
        {
            for (name, node) in old_children.iter() {
                match children.entry(name.clone()) {
                    Entry::Vacant(entry) => {
                        entry.insert(node.clone());
                    }
                    Entry::Occupied(entry) => {
                        if entry.get() == &TreeNode::Deleted {
                            // We don't insert the old node but we do remove the 'deleted' marker
                            // node!
                            entry.remove();
                        } else {
                            integrate_node_in_place(entry.into_mut(), node)?;
                        }
                    }
                }
            }
        }
    } else {
        // the node stays the same...
        // intentional NOP!
    }
    Ok(())
}

/// Given a node, recursively constructs a UID and GID lookup table based on THIS system's
/// users and groups.
pub fn create_uidgid_lookup_tables(
    node: &TreeNode,
    uids: &mut BTreeMap<u16, Option<String>>,
    gids: &mut BTreeMap<u16, Option<String>>,
) -> anyhow::Result<()> {
    let ownership = match &node {
        TreeNode::NormalFile { ownership, .. } => Some(ownership),
        TreeNode::Directory { ownership, .. } => Some(ownership),
        TreeNode::SymbolicLink { ownership, .. } => Some(ownership),
        TreeNode::Deleted => None,
    };
    if let Some(ownership) = ownership {
        if !uids.contains_key(&ownership.uid) {
            if let Some(user) = users::get_user_by_uid(ownership.uid.into()) {
                uids.insert(
                    ownership.uid,
                    Some(
                        user.name()
                            .to_str()
                            .ok_or(anyhow!("uid leads to non-String name"))?
                            .to_owned(),
                    ),
                );
            } else {
                uids.insert(ownership.uid, None);
            }
        }
        if !gids.contains_key(&ownership.gid) {
            if let Some(group) = users::get_group_by_gid(ownership.gid.into()) {
                gids.insert(
                    ownership.gid,
                    Some(
                        group
                            .name()
                            .to_str()
                            .ok_or(anyhow!("gid leads to non-String name"))?
                            .to_owned(),
                    ),
                );
            } else {
                gids.insert(ownership.gid, None);
            }
        }
    }

    if let TreeNode::Directory { children, .. } = &node {
        for (_name, child) in children {
            create_uidgid_lookup_tables(child, uids, gids)?;
        }
    }

    Ok(())
}

#[must_use]
/// Accepts a target UID/GID map and a source UID/GID map.
/// Merges source into target.
/// Returns true if everything is OK, false if there is a warning (overlap!).
/// In the event of an overlap, the pre-existing entry wins.
pub fn merge_uid_or_gid_tables(
    target: &mut BTreeMap<u16, Option<String>>,
    source: &BTreeMap<u16, Option<String>>,
) -> bool {
    let mut warn = false;
    for (uid_or_gid, name) in source.iter() {
        match target.entry(*uid_or_gid) {
            Entry::Vacant(entry) => {
                entry.insert(name.clone());
            }
            Entry::Occupied(entry) => {
                if entry.get() != name {
                    // Conflicting IDs
                    warn = true;
                }
            }
        }
    }
    !warn
}

#[cfg(test)]
mod tests {
    use crate::definitions::{
        FilesystemOwnership, FilesystemPermissions, RecursiveChunkRef, TreeNode,
    };
    use crate::tree::{differentiate_node_in_place, integrate_node_in_place};
    use std::collections::BTreeMap;

    #[test]
    pub fn test_differentiate_in_place_primitive() {
        let mut new = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 42, gid: 47 },
            permissions: FilesystemPermissions { mode: 1337 },
            children: BTreeMap::new(),
        };
        let old = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 42, gid: 47 },
            permissions: FilesystemPermissions { mode: 1337 },
            children: BTreeMap::new(),
        };
        assert!(differentiate_node_in_place(&mut new, &old).is_ok());

        assert_eq!(
            new,
            TreeNode::Directory {
                ownership: FilesystemOwnership { uid: 42, gid: 47 },
                permissions: FilesystemPermissions { mode: 1337 },
                children: BTreeMap::new()
            }
        );
    }

    /// Tests the creation of a file in the tree.
    #[test]
    pub fn test_differentiate_in_place_create_only() {
        let alice = TreeNode::NormalFile {
            mtime: 98347523,
            ownership: FilesystemOwnership { uid: 43, gid: 48 },
            permissions: FilesystemPermissions { mode: 1338 },
            content: RecursiveChunkRef {
                chunk_id: [36; 32],
                depth: 22,
            },
        };
        let bob_new = TreeNode::SymbolicLink {
            ownership: FilesystemOwnership { uid: 43, gid: 48 },
            target: "alice".to_string(),
        };

        let mut children_new = BTreeMap::new();
        children_new.insert("bob".to_owned(), bob_new.clone());
        children_new.insert("alice".to_owned(), alice.clone());

        let mut children_old = BTreeMap::new();
        children_old.insert("alice".to_owned(), alice.clone());

        let mut new = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 42, gid: 47 },
            permissions: FilesystemPermissions { mode: 1337 },
            children: children_new.clone(),
        };
        let old = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 41, gid: 46 },
            permissions: FilesystemPermissions { mode: 1336 },
            children: children_old.clone(),
        };

        let mut children_result = BTreeMap::new();
        children_result.insert("bob".to_owned(), bob_new);

        assert!(differentiate_node_in_place(&mut new, &old).is_ok());
        assert_eq!(
            new,
            TreeNode::Directory {
                ownership: FilesystemOwnership { uid: 42, gid: 47 },
                permissions: FilesystemPermissions { mode: 1337 },
                children: children_result
            }
        );
    }

    /// Tests only a change in metadata in the tree.
    #[test]
    pub fn test_differentiate_in_place_meta_only() {
        let alice = TreeNode::NormalFile {
            mtime: 98347523,
            ownership: FilesystemOwnership { uid: 43, gid: 48 },
            permissions: FilesystemPermissions { mode: 1338 },
            content: RecursiveChunkRef {
                chunk_id: [36; 32],
                depth: 22,
            },
        };

        let mut children = BTreeMap::new();
        children.insert("alice".to_owned(), alice);

        let mut new = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 42, gid: 47 },
            permissions: FilesystemPermissions { mode: 1337 },
            children: children.clone(),
        };
        let old = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 41, gid: 46 },
            permissions: FilesystemPermissions { mode: 1336 },
            children: children.clone(),
        };

        assert!(differentiate_node_in_place(&mut new, &old).is_ok());
        assert_eq!(
            new,
            TreeNode::Directory {
                ownership: FilesystemOwnership { uid: 42, gid: 47 },
                permissions: FilesystemPermissions { mode: 1337 },
                children: BTreeMap::new()
            }
        );
    }

    /// Tests that nodes that vanish lead to creation of a Deleted node.
    #[test]
    pub fn test_differences_in_place_deleted() {
        let alice = TreeNode::NormalFile {
            mtime: 98347523,
            ownership: FilesystemOwnership { uid: 43, gid: 48 },
            permissions: FilesystemPermissions { mode: 1338 },
            content: RecursiveChunkRef {
                chunk_id: [36; 32],
                depth: 22,
            },
        };
        let bob_old = TreeNode::SymbolicLink {
            ownership: FilesystemOwnership { uid: 43, gid: 48 },
            target: "alice".to_string(),
        };

        let mut children_old = BTreeMap::new();
        children_old.insert("bob".to_owned(), bob_old.clone());
        children_old.insert("alice".to_owned(), alice.clone());

        let mut children_new = BTreeMap::new();
        children_new.insert("alice".to_owned(), alice.clone());

        let old = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 42, gid: 47 },
            permissions: FilesystemPermissions { mode: 1337 },
            children: children_old.clone(),
        };
        let mut new = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 41, gid: 46 },
            permissions: FilesystemPermissions { mode: 1336 },
            children: children_new.clone(),
        };

        let mut children_result = BTreeMap::new();
        children_result.insert("bob".to_owned(), TreeNode::Deleted);

        assert!(differentiate_node_in_place(&mut new, &old).is_ok());
        assert_eq!(
            new,
            TreeNode::Directory {
                ownership: FilesystemOwnership { uid: 41, gid: 46 },
                permissions: FilesystemPermissions { mode: 1336 },
                children: children_result
            }
        );
    }

    #[test]
    pub fn test_differentiate_node_in_place_mega_example() {
        // TODO extend this example
        let parent = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0550 },
                    children: vec![
                        (
                            "file1".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1996,
                                ownership: FilesystemOwnership { uid: 54, gid: 59 },
                                permissions: FilesystemPermissions { mode: 0311 },
                                content: RecursiveChunkRef {
                                    chunk_id: [37; 32],
                                    depth: 2,
                                },
                            },
                        ),
                        (
                            "file2".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                content: RecursiveChunkRef {
                                    chunk_id: [42; 32],
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        let child_full = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0560 },
                    children: vec![
                        (
                            "file1".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1996,
                                ownership: FilesystemOwnership { uid: 54, gid: 59 },
                                permissions: FilesystemPermissions { mode: 0311 },
                                content: RecursiveChunkRef {
                                    chunk_id: [37; 32],
                                    depth: 2,
                                },
                            },
                        ),
                        (
                            "file42".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                content: RecursiveChunkRef {
                                    chunk_id: [42; 32],
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        let mut child_diff = child_full.clone();
        differentiate_node_in_place(&mut child_diff, &parent).unwrap();

        let expected_child_diff = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0560 },
                    children: vec![
                        ("file2".to_string(), TreeNode::Deleted),
                        (
                            "file42".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                content: RecursiveChunkRef {
                                    chunk_id: [42; 32],
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        assert_eq!(child_diff, expected_child_diff);
    }

    #[test]
    pub fn test_integrate_node_in_place_mega_example() {
        // TODO extend this example
        let parent = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0550 },
                    children: vec![
                        (
                            "file1".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1996,
                                ownership: FilesystemOwnership { uid: 54, gid: 59 },
                                permissions: FilesystemPermissions { mode: 0311 },
                                content: RecursiveChunkRef {
                                    chunk_id: [37; 32],
                                    depth: 2,
                                },
                            },
                        ),
                        (
                            "file2".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                content: RecursiveChunkRef {
                                    chunk_id: [42; 32],
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        let child_diff = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0560 },
                    children: vec![
                        ("file2".to_string(), TreeNode::Deleted),
                        (
                            "file42".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                content: RecursiveChunkRef {
                                    chunk_id: [42; 32],
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        let mut child_full = child_diff.clone();
        integrate_node_in_place(&mut child_full, &parent).unwrap();

        let expected_child_full = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0560 },
                    children: vec![
                        (
                            "file1".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1996,
                                ownership: FilesystemOwnership { uid: 54, gid: 59 },
                                permissions: FilesystemPermissions { mode: 0311 },
                                content: RecursiveChunkRef {
                                    chunk_id: [37; 32],
                                    depth: 2,
                                },
                            },
                        ),
                        (
                            "file42".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                content: RecursiveChunkRef {
                                    chunk_id: [42; 32],
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        assert_eq!(child_full, expected_child_full);
    }
}
